from copy import copy
from itertools import compress
from typing import Generator, List

from hwt.doc_markers import internal
from hwt.hdl.statements.codeBlockContainer import HdlStmCodeBlockContainer
from hwt.hdl.statements.statement import HwtSyntaxError, HdlStatement
from hwt.pyUtils.uniqList import UniqList
from hwt.synthesizer.rtlLevel.constants import NOT_SPECIFIED
from hwt.synthesizer.rtlLevel.fill_stm_list_with_enclosure import fill_stm_list_with_enclosure, \
    HdlAssignmentContainer_constructor
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.synthesizer.rtlLevel.reduce_processes import reduceProcesses
from hwt.synthesizer.rtlLevel.rtlSignal import RtlSignal


@internal
def cut_off_drivers_of(dstSignal: RtlSignal, statements: List[HdlStatement]):
    """
    Cut off drivers from statements
    """
    separated = []
    stm_filter = []
    for stm in statements:
        stm._clean_signal_meta()
        d = stm._cut_off_drivers_of(dstSignal)
        if d is not None:
            separated.append(d)

        f = d is not stm
        stm_filter.append(f)

    return list(compress(statements, stm_filter)), separated


@internal
def name_for_process(outputs: List[RtlSignal]) -> str:
    """
    Resolve name for process
    """
    out_names = []
    for sig in outputs:
        if not sig.hasGenericName:
            out_names.append(sig.name)

    if out_names:
        return min(out_names)
    else:
        return ""


@internal
def _statements_to_HdlStmCodeBlockContainers(_statements, tryToSolveCombLoops)\
        ->Generator[HdlStmCodeBlockContainer, None, None]:
    assert _statements
    # try to simplify statements
    proc_statements = []
    for _stm in _statements:
        _stm._clean_signal_meta()
        stms, _ = _stm._try_reduce()
        proc_statements.extend(stms)

    if not proc_statements:
        return

    outputs = UniqList()
    _inputs = UniqList()
    sensitivity = UniqList()
    enclosed_for = set()
    _proc_statements = []
    for _stm in proc_statements:
        seen = set()
        _stm._discover_sensitivity(seen)
        _stm._discover_enclosure()
        if _stm._outputs:
            # remove a statement entirely if it has no ouput
            # (empty if statment or something similar)
            # simulation only processes should not be processed by this function
            # and process should always drive something, unless it is useless
            outputs.extend(_stm._outputs)
            _inputs.extend(_stm._inputs)
            sensitivity.extend(_stm._sensitivity)
            enclosed_for.update(_stm._enclosed_for)
            _proc_statements.append(_stm)

    proc_statements = _proc_statements
    if not proc_statements:
        # this can happen e.g. when If does not contains any HdlAssignmentContainer
        return
    sensitivity_recompute = False
    enclosure_recompute = False
    enclosure_values = {}
    for sig in outputs:
        # inject nop_val if needed
        if sig._nop_val is not NOT_SPECIFIED and sig not in enclosed_for:
            enclosure_recompute = True
            n = sig._nop_val
            enclosure_values[sig] = HdlAssignmentContainer_constructor(n, sig)
            if isinstance(n, RtlSignalBase):
                _inputs.append(n)
                sensitivity_recompute = True

    if enclosure_recompute:
        # we have some enclosure values, try fill missing code branches with
        # this values
        do_enclose_for = [o for o in outputs if o in enclosure_values]
        fill_stm_list_with_enclosure(None, enclosed_for, proc_statements,
                                     do_enclose_for, enclosure_values)

    if enclosure_recompute or sensitivity_recompute:
        for _stm in proc_statements:
            _stm._clean_signal_meta()
            seen = set()
            _stm._discover_sensitivity(seen)
            _stm._discover_enclosure()

        if sensitivity_recompute:
            sensitivity.clear()
            for _stm in proc_statements:
                sensitivity.extend(_stm._sensitivity)

    for o in outputs:
        assert not o.hidden, o

    seen = set()
    inputs = UniqList()
    for i in _inputs:
        inputs.extend(i._walk_public_drivers(seen))

    intersect = outputs.intersection_set(sensitivity)
    if intersect:
        # there is a combinational loop inside a single process which
        # can not be solved by separation of statments in process
        if not tryToSolveCombLoops:
            raise HwtSyntaxError(
                "Combinational loop on signal(s)", intersect)

        # try to solve combinational loops by separating drivers of signals
        # from statements
        for sig in intersect:
            proc_statements, proc_stms_select = cut_off_drivers_of(
                sig, proc_statements)
            yield from _statements_to_HdlStmCodeBlockContainers(proc_stms_select, False)

        if proc_statements:
            yield from _statements_to_HdlStmCodeBlockContainers(proc_statements, False)
    else:
        # no combinational loops, wrap current statemetns to a process instance

        name = name_for_process(outputs)
        yield HdlStmCodeBlockContainer.from_known_io(
            "assig_process_" + name,
            proc_statements, sensitivity,
            inputs, outputs)


@internal
def statements_to_HdlStmCodeBlockContainers(statements: List[HdlStatement])\
        ->Generator[HdlStmCodeBlockContainer, None, None]:
    """
    Pack statements into HdlStmCodeBlockContainer instances

    * for each out signal resolve it's drivers and collect them
    * split statements if there is and combinational loop
    * merge statements if it is possible
    * resolve sensitivity lists
    * wrap into HdlStmCodeBlockContainer instance
    * for every IO of process generate name if signal has not any
    """
    # create copy because this set will be reduced
    statements = copy(statements)

    # process ranks = how many assignments is probably in process
    # used to minimize number of merge tries
    processes = []
    while statements:
        stm = statements.pop()
        proc_statements = [stm, ]
        ps = _statements_to_HdlStmCodeBlockContainers(proc_statements, True)
        processes.extend(ps)

    yield from reduceProcesses(processes)



from hdlConvertorAst.hdlAst import HdlValueId, HdlValueInt, HdlOp, \
    HdlOpType
from hdlConvertorAst.to.hdlUtils import bit_string
from hdlConvertorAst.translate.common.name_scope import LanguageKeyword
from hwt.hdl.operator import Operator
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.bitsVal import BitsVal
from hwt.hdl.types.defs import BOOL, BIT
from hwt.hdl.types.enumVal import HEnumVal
from hwt.hdl.types.sliceVal import HSliceVal
from hwt.hdl.value import HValue
from hwt.serializer.generic.value import ToHdlAst_Value
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


class ToHdlAstVhdl2008_Value(ToHdlAst_Value):

    TRUE = HdlValueId("TRUE", obj=LanguageKeyword())
    FALSE = HdlValueId("FALSE", obj=LanguageKeyword())
    #TO_UNSIGNED = HdlValueId("TO_UNSIGNED", obj=LanguageKeyword())
    #TO_SIGNED = HdlValueId("TO_SIGNED", obj=LanguageKeyword())

    def as_hdl_cond(self, c, forceBool):
        assert isinstance(c, (RtlSignalBase, HValue)), c
        if not forceBool or c._dtype == BOOL:
            return self.as_hdl(c)
        elif c._dtype == BIT:
            return self.as_hdl(c._eq(1))
        elif isinstance(c._dtype, Bits):
            return self.as_hdl(c != 0)
        else:
            raise NotImplementedError()

    def as_hdl_HEnumVal(self, val: HEnumVal):
        name = self.name_scope.get_object_name(val)
        return HdlValueId(name, obj=val)

    def as_hdl_HArrayVal(self, val):
        return [self.as_hdl_Value(v) for v in val]

    def sensitivityListItem(self, item, anyIsEventDependnt):
        if isinstance(item, Operator):
            item = item.operands[0]
        return self.as_hdl(item)

    def as_hdl_BitString(self, v, width: int,
                         force_vector: bool, vld_mask: int, signed):
        is_bit = not force_vector and width == 1
        #if vld_mask != mask(width) or width >= 32 or is_bit:
        v = bit_string(v, width, vld_mask)
        if is_bit:
            v.base = 256
            return v
        if signed is None:
            return v
        elif signed:
            cast = self.SIGNED
        else:
            cast = self.UNSIGNED
        return HdlOp(HdlOpType.APOSTROPHE, [cast, v])

        #else:
        #    v = HdlValueInt(v, None, None)
        #
        #    if signed is None:
        #        return v
        #    elif signed:
        #        cast_fn = self.TO_SIGNED
        #    else:
        #        cast_fn = self.TO_UNSIGNED
        #    return hdl_call(cast_fn, [v, HdlValueInt(width, None, None)])

    def as_hdl_BoolVal(self, val: BitsVal):
        if val.val:
            return self.TRUE
        else:
            return self.FALSE

    def as_hdl_BitsVal(self, val: BitsVal):
        t = val._dtype
        v = super(ToHdlAstVhdl2008_Value, self).as_hdl_BitsVal(val)
        # handle '1' vs "1" difference (bit literal vs vector)
        if not t.force_vector and t.bit_length() == 1 and t != BOOL:
            if isinstance(v, HdlValueInt):
                v.base = 256
            else:
                # assert is cast
                assert isinstance(v, HdlOp) and v.fn == HdlOpType.CALL, v
                _v = v.ops[1]
                if isinstance(_v, HdlValueInt):
                    _v.base = 256
                else:
                    raise NotImplementedError()
        return v

    def as_hdl_HSliceVal(self, val: HSliceVal):
        upper = val.val.start
        if int(val.val.step) == -1:
            if isinstance(upper, HValue):
                upper = HdlValueInt(int(upper) - 1, None, None)
            else:
                upper = HdlOp(HdlOpType.SUB, [self.as_hdl_Value(upper),
                                                   HdlValueInt(1, None, None)])
        else:
            raise NotImplementedError(val.val.step)

        return HdlOp(HdlOpType.DOWNTO, [upper, self.as_hdl(val.val.stop)])



from hwt.hdl.value import HValue
from hwt.hdl.types.defs import BOOL
from hwt.hdl.types.typeCast import toHVal
from hwt.doc_markers import internal


class HStringVal(HValue):
    """
    Value class for hdl HString type
    """

    @classmethod
    def from_py(cls, typeObj, val, vld_mask=None):
        """
        :param val: python string or None
        :param typeObj: instance of HString HdlType
        :param vld_mask: if is None validity is resolved from val
            if is 0 value is invalidated
            if is 1 value has to be valid
        """
        assert isinstance(val, str) or val is None
        vld = 0 if val is None else 1
        if not vld:
            assert vld_mask is None or vld_mask == 0
            val = ""
        else:
            if vld_mask == 0:
                val = ""
                vld = 0

        return cls(typeObj, val, vld)

    def to_py(self):
        if not self._is_full_valid():
            raise ValueError(f"Value of {self} is not fully defined")
        return self.val

    @internal
    def _eq__val(self, other):
        eq = self.val == other.val
        vld = int(self.vld_mask and other.vld_mask)

        return BOOL.getValueCls()(BOOL, int(eq), vld)

    def _eq(self, other):
        other = toHVal(other, self._dtype)
        if isinstance(other, HValue):
            return self._eq__val(other)
        else:
            raise NotImplementedError()



"""
Hwt serializer converts  HDL objects back to code in python for hwt.
"""
from hdlConvertorAst.to.hwt._main import ToHwt
from hwt.serializer.hwt.serializer import ToHdlAstHwt
from hwt.serializer.serializer_config import DummySerializerConfig


class HwtSerializer(DummySerializerConfig):
    fileExtension = '.py'
    TO_HDL_AST = ToHdlAstHwt
    TO_HDL = ToHwt



from typing import Generator


def _tarjan_head(index, lowlink, S, S_set, T, g, v):
    lowlink[v] = index[v] = len(index)
    S.append(v)
    S_set.add(v)
    it = iter(g.get(v, ()))
    T.append((it, False, v, None))


class StronglyConnectedComponentSearchTarjan():
    """
    Tarjan's strongly connected component search graph algorithm for DAGs
    
    based on https://github.com/bwesterb/py-tarjan
    """

    def __init__(self, g: dict):
        """
        :param g: graph represented as a dictionary { <vertex> : <successors of vertex> }.
        :note: vertex type does not matter but it has to be hashable (implement __eq__ and __hash__)
        """
        self.g = g

    def search_strongly_connected_components(self) -> Generator:
        """
        yields the strongly connected components of the graph in a topological order.
        """
        g = self.g # the graph
        S = [] # The main stack of the alg.
        S_set = set() # == set(S) for performance
        index = {} # { v : <index of v> }
        lowlink = {} # { v : <lowlink of v> }
        T = [] # stack to replace recursion
        main_iter = iter(g)
        for v in main_iter:
            if v not in index:
                _tarjan_head(index, lowlink, S, S_set, T, g, v)

            while T:
                it, inside, v, w = T.pop()
                if inside:
                    lowlink[v] = min(lowlink[w], lowlink[v])

                body_break = False
                for w in it:
                    if w not in index:
                        T.append((it, True, v, w))
                        _tarjan_head(index, lowlink, S, S_set, T, g, w)
                        body_break = True
                        break

                    if w in S_set:
                        lowlink[v] = min(lowlink[v], index[w])

                if body_break:
                    continue

                if lowlink[v] == index[v]:
                    scc = []
                    w = None
                    while v != w:
                        w = S.pop()
                        scc.append(w)
                        S_set.remove(w)

                    yield scc



from collections import deque

from hwt.hdl.constants import NOP
from hwt.simulator.agentBase import SyncAgentBase
from hwtSimApi.hdlSimulator import HdlSimulator
from hwtSimApi.triggers import WaitCombRead, WaitWriteOnly


class RdSyncedAgent(SyncAgentBase):
    """
    Simulation/verification agent for RdSynced interface
    """
    def __init__(self, sim: HdlSimulator, intf, allowNoReset=True):
        super().__init__(sim, intf, allowNoReset=allowNoReset)
        self.actualData = NOP
        self.data = deque()
        self._rd = self.get_ready_signal(intf)

    @classmethod
    def get_ready_signal(cls, intf):
        return intf.rd

    def get_ready(self):
        return self._rd.read()

    def set_ready(self, val):
        self._rd.write(val)

    def setEnable_asMonitor(self, en):
        super(RdSyncedAgent, self).setEnable_asMonitor(en)
        if not en:
            self.set_ready(0)

    def monitor(self):
        """Collect data from interface"""
        yield WaitCombRead()
        if self.notReset() and self._enabled:
            yield WaitWriteOnly()
            self.set_ready(1)

            yield WaitCombRead()
            d = self.get_data()
            self.data.append(d)
        else:
            yield WaitWriteOnly()
            self.set_ready(0)

    def get_data(self):
        """extract data from interface"""
        return self.intf.data.read()

    def set_data(self, data):
        """write data to interface"""
        self.intf.data.write(data)

    def driver(self):
        """Push data to interface"""
        yield WaitWriteOnly()
        if self.actualData is NOP and self.data:
            self.actualData = self.data.popleft()

        do = self.actualData is not NOP

        if do:
            self.set_data(self.actualData)
        else:
            self.set_data(None)

        yield WaitCombRead()
        en = self.notReset() and self._enabled
        if not (en and do):
            return

        if en:
            rd = self.get_ready()
            try:
                rd = int(rd)
            except ValueError:
                raise AssertionError(
                    ("%r: ready signal for interface %r is in invalid state,"
                     " this would cause desynchronization") %
                    (self.sim.now, self.intf))
            if rd:
                if self._debugOutput is not None:
                    self._debugOutput.write("%s, wrote, %d: %r\n" % (
                                               self.intf._getFullName(),
                                               self.sim.now, self.actualData))
                if self.data:
                    self.actualData = self.data.popleft()
                else:
                    self.actualData = NOP



from io import StringIO
import math
from typing import List, Tuple, Union

from hdlConvertorAst.hdlAst import HdlValueId
from hdlConvertorAst.hdlAst._defs import HdlIdDef
from hdlConvertorAst.to.vhdl.vhdl2008 import ToVhdl2008
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model.utils import hdl_call
from hwt.doc_markers import internal
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.defs import BOOL, STR, BIT, INT
from hwt.hdl.types.hdlType import HdlType
from hwt.serializer.store_manager import SaveToFilesFlat
from hwt.serializer.vhdl import Vhdl2008Serializer, ToHdlAstVhdl2008
from hwt.synthesizer.dummyPlatform import DummyPlatform
from hwt.synthesizer.interface import Interface
from hwt.synthesizer.interfaceLevel.unitImplHelpers import getSignalName
from hwt.synthesizer.param import Param
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.synthesizer.rtlLevel.rtlSignal import RtlSignal
from hwt.synthesizer.unit import Unit
from hwt.synthesizer.utils import to_rtl
from ipCorePackager.intfIpMeta import VALUE_RESOLVE
from ipCorePackager.otherXmlObjs import Value
from ipCorePackager.packager import IpCorePackager


class ToHdlAstVivadoTclExpr(ToHdlAstVhdl2008):
    _spirit_decode = HdlValueId("spirit:decode")
    _id = HdlValueId("id")

    def as_hdl_SignalItem(self, si, declaration=False):
        assert(declaration == False)
        if si.hidden:
            assert si.origin is not None, si
            return self.as_hdl(si.origin)
        else:
            id_ = hdl_call(self._id, [f'MODELPARAM_VALUE.{si.name}'])
            return hdl_call(self._spirit_decode, [id_])


class IpPackager(IpCorePackager):
    """
    IP-core packager

    :summary: Packs HDL, constraint and other files to IP-Core package
        for distribution and simple integration

    """

    def __init__(self, topUnit: Unit, name: str=None,
                 extra_files: List[str]=[],
                 serializer_cls=Vhdl2008Serializer,
                 target_platform=DummyPlatform()):
        """
        :param topObj: :class:`hwt.synthesizer.unit.Unit` instance of top component
        :param name: optional name of top
        :param extra_files: list of extra HDL/constrain file names for files
            which should be distributed in this IP-core
            (\*.v - verilog, \*.sv,\*.svh -system verilog, \*.vhd - vhdl, \*.xdc - XDC)
        :param serializer: serializer which specifies target HDL language
        :param target_platform: specifies properties of target platform, like available resources, vendor, etc.
        """
        if not name:
            name = topUnit._getDefaultName()

        super(IpPackager, self).__init__(
            topUnit, name, extra_files)
        self.serializer = serializer_cls
        self.target_platform = target_platform

    @internal
    def toHdlConversion(self, top, topName: str, saveTo: str) -> List[str]:
        """
        :param top: object which is represenation of design
        :param topName: name which should be used for ipcore
        :param saveTo: path of directory where generated files should be stored

        :return: list of file namens in correct compile order
        """
        ser = self.serializer
        store = SaveToFilesFlat(ser, saveTo)
        to_rtl(top, name=topName, store_manager=store,
              target_platform=self.target_platform)
        return store.files

    @internal
    def paramToIpValue(self, idPrefix: str, g: HdlIdDef, resolve) -> Value:
        val = Value()
        val.id = idPrefix + g.name
        if resolve is not VALUE_RESOLVE.NONE:
            val.resolve = resolve
        v = g.value
        t = g.type

        def getVal():
            if v.vld_mask:
                return v.val
            else:
                return 0

        def bitString(w):
            val.format = "bitString"
            digits = math.ceil(w / 4)
            val.text = ('0x%0' + str(digits) + 'X') % getVal()
            val.bitStringLength = str(w)

        if t == BOOL:
            val.format = "bool"
            val.text = str(bool(getVal())).lower()
        elif t == INT:
            val.format = "long"
            val.text = str(getVal())
        elif t == STR:
            val.format = "string"
            val.text = v.val
        elif isinstance(t, Bits):
            bitString(t.bit_length())
        else:
            raise NotImplementedError(
                f"Not implemented for datatype {t}")
        return val

    @internal
    def getParamPhysicalName(self, p: HdlIdDef):
        return p.name

    @internal
    def getParamType(self, p: HdlIdDef) -> HdlType:
        assert p.type in [INT, BOOL, STR], p
        return p.type

    @internal
    def iterParams(self, unit: Unit):
        return unit._ctx.ent.params

    @internal
    def iterInterfaces(self, top: Unit):
        return top._interfaces

    @internal
    def serializeType(self, hdlType: HdlType) -> str:
        """
        :see: doc of method on parent class
        """
        buff = StringIO()
        to_ast = ToHdlAstVhdl2008()
        hdl = to_ast.as_hdl_HdlType(hdlType)
        ser = ToVhdl2008(buff)
        ser.visit_iHdlObj(hdl)
        return buff.getvalue()

    @internal
    def getVectorFromType(self, dtype) -> Union[bool, None, Tuple[int, int]]:
        """
        :see: doc of method on parent class
        """
        if dtype == BIT:
            return False
        elif isinstance(dtype, Bits):
            return [dtype.bit_length() - 1, INT.from_py(0)]

    @internal
    def getInterfaceType(self, intf: Interface) -> HdlType:
        """
        :see: doc of method on parent class
        """
        return intf._dtype

    @internal
    def getInterfaceLogicalName(self, intf: Interface):
        """
        :see: doc of method on parent class
        """
        return getSignalName(intf)

    @internal
    def getInterfacePhysicalName(self, intf: Interface):
        """
        :see: doc of method on parent class
        """
        return intf._sigInside.name

    @internal
    def getInterfaceDirection(self, thisIntf):
        """
        :see: doc of method on parent class
        """
        return thisIntf._direction

    @internal
    def getTypeWidth(self, dtype: HdlType, do_eval=False)\
            -> Tuple[int, Union[int, RtlSignal], bool]:
        """
        :see: doc of method on parent class
        """
        width = dtype.bit_length()
        widthStr = str(width)
        return width, widthStr, False

    @internal
    def getObjDebugName(self, obj: Union[Interface, Unit, Param]) -> str:
        """
        :see: doc of method on parent class
        """
        return obj._getFullName()

    @internal
    def serialzeValueToTCL(self, val, do_eval=False) -> Tuple[str, str, bool]:
        """
        :see: doc of method on parent class
        """
        if isinstance(val, int):
            val = INT.from_py(val)
        if do_eval:
            val = val.staticEval()

        buff = StringIO()
        to_hdl = ToHdlAstVivadoTclExpr()
        ser = Vhdl2008Serializer.TO_HDL(buff)

        hdl = to_hdl.as_hdl(val)
        ser.visit_iHdlObj(hdl)
        tclVal = buff.getvalue()

        if isinstance(val, RtlSignalBase):
            buff = StringIO()
            hdl = to_hdl.as_hdl(val.staticEval())
            ser = Vhdl2008Serializer.TO_HDL(buff)
            ser.visit_iHdlObj(hdl)
            tclValVal = buff.getvalue()

            return tclVal, tclValVal, False
        else:
            return tclVal, tclVal, True



"""
This module contains the objects to store hardware constraints.
Hardware constrains are usually stored in XDC/UCF files and they specify somethings
which can not be described using HDL (SystemVerilog/VHDL) like relation between clock.
Placement of component if FPGA etc.
"""

from copy import copy
from typing import Union, Tuple

from hwt.synthesizer.interface import Interface
from hwt.synthesizer.rtlLevel.rtlSignal import RtlSignal
from hwt.synthesizer.unit import Unit
from hwt.synthesizer.componentPath import ComponentPath


class iHdlConstrain():

    def _get_parent(self) -> Unit:
        raise NotImplementedError(self)

    def _copy_with_root_upadate(self, old_path_prefix, new_path_prefix):
        raise NotImplementedError()

    def register_on_parent(self):
        self._get_parent()._constraints.append(self)


def _get_parent_unit(path: Tuple[Union[Unit, Interface, RtlSignal, iHdlConstrain], ...]) -> Unit:
    """
    Search parent :class:`hwt.synthesizer.unit.Unit` instance in path
    """
    if isinstance(path, iHdlConstrain):
        return path._get_parent()

    for o in reversed(path):
        if isinstance(o, Unit):
            return o

    raise AssertionError("No parent unit in path", path)


def _get_absolute_path(obj) -> Union[Tuple[Union[Unit, Interface, RtlSignal, iHdlConstrain], ...], None]:
    """
    Get tuple containing a path of objects from top to this object
    """
    if obj is None:
        return None
    elif isinstance(obj, iHdlConstrain):
        return obj

    return ComponentPath(obj).resolve()


def _apply_path_update(path: ComponentPath, old_path_prefix: ComponentPath, new_path_prefix: ComponentPath):
    """
    Update prefix of the path tuple
    """
    if isinstance(path, iHdlConstrain):
        return path._copy_with_root_upadate(old_path_prefix, new_path_prefix)
    return path.update_prefix(old_path_prefix, new_path_prefix)


class set_max_delay(iHdlConstrain):
    """
    Object which represents the max_delay constrain

    * usually used to set propagation time between two clock domains etc.

    :ivar ~.start: start of the signal path
    :ivar ~.end: end of the signal path
    :ivar ~.time_ns: max delay of the specified path in ns
    """

    def __init__(self,
                 start: Union[Interface, RtlSignal],
                 end: Union[Interface, RtlSignal],
                 time_ns: float,
                 datapath_only=True,
                 ommit_registration=False):
        self.start = _get_absolute_path(start)
        self.end = _get_absolute_path(end)
        self.time_ns = time_ns
        self.datapath_only = datapath_only
        if not ommit_registration:
            self.register_on_parent()

    def _copy_with_root_upadate(self, old_path_prefix, new_path_prefix):
        new_o = copy(self)
        new_o.start = _apply_path_update(
            self.start, old_path_prefix, new_path_prefix)
        new_o.end = _apply_path_update(
            self.end, old_path_prefix, new_path_prefix)
        return new_o

    def _get_parent(self) -> Unit:
        return _get_parent_unit(self.end)


class set_false_path(iHdlConstrain):

    def __init__(self, start: Union[None, Interface, RtlSignal],
                 end: Union[None, Interface, RtlSignal],
                 ommit_registration=False):
        self.start = _get_absolute_path(start)
        self.end = _get_absolute_path(end)
        if not ommit_registration:
            self.register_on_parent()

    def _copy_with_root_upadate(self, old_path_prefix, new_path_prefix):
        return set_max_delay._copy_with_root_upadate(self, old_path_prefix, new_path_prefix)

    def _get_parent(self) -> Unit:
        o = self.start
        if o is None:
            o = self.end
        return _get_parent_unit(o)


class get_clock_of(iHdlConstrain):

    def __init__(self, obj: Union[Interface, RtlSignal],
                 ommit_registration=False):
        self.obj = _get_absolute_path(obj)

    def _copy_with_root_upadate(self, old_path_prefix, new_path_prefix):
        new_o = copy(self)
        new_o.obj = _apply_path_update(
            self.obj, old_path_prefix, new_path_prefix)
        return new_o

    def _get_parent(self) -> Unit:
        return _get_parent_unit(self.obj)


class set_async_reg(iHdlConstrain):
    """
    Placement constrain which tell that the register should be put as close as possible to it's src/dst

    It should not be placed on the FF on the src domain,
    but should be set on FFs (possibly more) on the destination domain.
    """

    def __init__(self, sig: RtlSignal,
                 ommit_registration=False):
        self.sig = _get_absolute_path(sig)
        if not ommit_registration:
            self.register_on_parent()

    def _copy_with_root_upadate(self, old_path_prefix, new_path_prefix):
        new_o = copy(self)
        new_o.sig = _apply_path_update(
            self.sig, old_path_prefix, new_path_prefix)
        return new_o

    def _get_parent(self) -> Unit:
        return _get_parent_unit(self.sig)




from functools import reduce
from math import ceil
from typing import Union

from hwt.doc_markers import internal
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.hdlType import HdlType
from hwt.hdl.types.utils import walkFlattenFields
from hwt.hdl.value import HValue
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


class BitWidthErr(Exception):
    """
    Wrong bit width of signal/value
    """


def fitTo_t(what: Union[RtlSignalBase, HValue], where_t: HdlType,
            extend: bool=True, shrink: bool=True):
    """
    Slice signal "what" to fit in "where"
    or
    arithmetically (for signed by MSB / unsigned, vector with 0) extend
    "what" to same width as "where"

    little-endian impl.

    :param extend: allow increasing of the signal width
    :param shrink: allow shrinking of the signal width
    """

    whatWidth = what._dtype.bit_length()
    toWidth = where_t.bit_length()
    if toWidth == whatWidth:
        return what
    elif toWidth < whatWidth:
        # slice
        if not shrink:
            raise BitWidthErr()

        return what[toWidth:]
    else:
        if not extend:
            raise BitWidthErr()

        w = toWidth - whatWidth

        if what._dtype.signed:
            # signed extension
            msb = what[whatWidth - 1]
            ext = reduce(lambda a, b: a._concat(b), [msb for _ in range(w)])
        else:
            # 0 extend
            ext = Bits(w).from_py(0)

        return ext._concat(what)


def fitTo(what: Union[RtlSignalBase, HValue], where: Union[RtlSignalBase, HValue],
          extend: bool=True, shrink: bool=True):
    return fitTo_t(what, where._dtype, extend, shrink)


class NotEnoughtBitsErr(Exception):
    """
    More bits is required for such an operation
    """


class BitWalker():
    """
    Walker which can walk chunks of bits on signals/values of all types

    :ivar ~.sigOrVal: signal or value to iterate over
    :ivar ~.fillup: flag that means that if there is not enough bits
        for last item fill it up with invalid bits (otherwise raise)
    """

    def __init__(self, sigOrVal: Union[RtlSignalBase, HValue],
                 skipPadding: bool=True,
                 fillup: bool=False):
        """
        :param skipPadding: if true padding is skipped in dense types
        """
        self.it = walkFlattenFields(sigOrVal, skipPadding=skipPadding)
        self.fillup = fillup
        self.actuallyHave = 0
        self.actual = None
        self.actualOffset = 0

    @internal
    def _get(self, numberOfBits: int, doCollect: bool):
        """
        :param numberOfBits: number of bits to get from actual position
        :param doCollect: if False output is not collected just iterator moves
            in data structure
        """
        if not isinstance(numberOfBits, int):
            numberOfBits = int(numberOfBits)

        while self.actuallyHave < numberOfBits:
            # accumulate while not has enough
            try:
                f = next(self.it)
            except StopIteration:
                if self.fillup and self.actual is not None:
                    break
                else:
                    raise NotEnoughtBitsErr()

            thisFieldLen = f._dtype.bit_length()
            if self.actual is None:
                if not doCollect and thisFieldLen <= numberOfBits:
                    numberOfBits -= thisFieldLen
                else:
                    self.actual = f
                    self.actuallyHave = thisFieldLen
            else:
                if not doCollect and self.actuallyHave < numberOfBits:
                    self.actuallyHave = thisFieldLen
                    self.actual = f
                else:
                    self.actuallyHave += thisFieldLen
                    self.actual = f._concat(self.actual)

        # slice out from actual
        actual = self.actual
        actualOffset = self.actualOffset

        if self.actuallyHave < numberOfBits:
            assert self.fillup
            if doCollect:
                t = self.actual._dtype
                fillupW = numberOfBits - self.actuallyHave
                padding_t = Bits(fillupW, signed=t.signed, negated=t.negated)
                padding = padding_t.from_py(None)
                actual = padding._concat(actual)
            self.actuallyHave = 0
            self.actualOffset = 0
        else:
            # update about what was taken
            self.actuallyHave -= numberOfBits
            self.actualOffset += numberOfBits

        if self.actuallyHave == 0:
            self.actual = None
            self.actualOffset = 0

        if doCollect:
            if numberOfBits == 1:
                return actual[actualOffset]
            else:
                return actual[(actualOffset + numberOfBits):actualOffset]

    def get(self, numberOfBits: int) -> Union[RtlSignalBase, HValue]:
        """
        :param numberOfBits: number of bits to get from actual position
        :return: chunk of bits of specified size (instance of Value or RtlSignal)
        """
        return self._get(numberOfBits, True)

    def skip(self, numberOfBits: int) -> None:
        """
        Move this iterator without care about item

        :param numberOfBits: number of bits to get from actual position
        """
        self._get(numberOfBits, False)

    def assertIsOnEnd(self):
        """
        Assert there is nothing left in this iterator
        """
        try:
            next(self.it)
        except StopIteration:
            return

        raise AssertionError("there are still some items")


def iterBits(sigOrVal: Union[RtlSignalBase, HValue], bitsInOne: int=1,
             skipPadding: bool=True, fillup: bool=False):
    """
    Iterate over bits in vector

    :param sigOrVal: signal or value to iterate over
    :param bitsInOne: number of bits in one part
    :param skipPadding: if true padding is skipped in dense types
    :param fillup: flag that means that if there is not enough bits
        for last item fill it up with invalid bits (otherwise raise)
    """
    bw = BitWalker(sigOrVal, skipPadding, fillup)
    try:
        bit_len = sigOrVal._dtype.bit_length()
    except TypeError:
        bit_len = None
    if bit_len is None:
        try:
            while True:
                yield bw.get(bitsInOne)
        except NotEnoughtBitsErr:
            return
    else:
        for _ in range(ceil(bit_len / bitsInOne)):
            yield bw.get(bitsInOne)

        bw.assertIsOnEnd()



from hdlConvertorAst.hdlAst._bases import iHdlStatement
from hdlConvertorAst.hdlAst._statements import HdlStmAssign, HdlStmCase, \
    HdlStmBlock, HdlStmBreak
from hdlConvertorAst.to.verilog.constants import SIGNAL_TYPE
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model.utils import hdl_getattr, \
    hdl_call
from hwt.doc_markers import internal
from hwt.hdl.statements.assignmentContainer import HdlAssignmentContainer
from hwt.hdl.operator import Operator
from hwt.hdl.operatorDefs import AllOps
from hwt.hdl.statements.switchContainer import SwitchContainer
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.defs import BOOL
from hwt.hdl.variables import SignalItem
from hwt.serializer.exceptions import SerializerException
from hwt.serializer.systemC.utils import systemCTypeOfSig
from hwt.serializer.verilog.value import ToHdlAstVerilog_Value


class ToHdlAstSystemC_statements():

    def has_to_be_process(self, proc: iHdlStatement):
        return True

    def can_pop_process_wrap(self, statements, hasToBeVhdlProcess):
        return False

    def sensitivityListItem(self, item, anyIsEventDependent):
        orig_in_sensitivity_list = self._in_sensitivity_list
        try:
            self._in_sensitivity_list = True
            return ToHdlAstVerilog_Value.sensitivityListItem(
                self, item, anyIsEventDependent)
        finally:
            self._in_sensitivity_list = orig_in_sensitivity_list

    def as_hdl_SwitchContainer(self, sw: SwitchContainer) -> HdlStmCase:
        """
        Same as parent as_hdl_SwitchContainer but add "break" to all cases
        """
        sw_hdl = super(ToHdlAstSystemC_statements, self).as_hdl_SwitchContainer(sw)
        new_cases = []
        for c, stm in sw_hdl.cases:
            if not isinstance(stm, HdlStmBlock):
                _stm = HdlStmBlock()
                _stm.body.append(stm)
                stm = _stm

            stm.body.append(HdlStmBreak())
            new_cases.append((c, stm))
        sw_hdl.cases = new_cases
        return sw_hdl

    @internal
    def _as_hdl_HdlAssignmentContainer(self, dst, typeOfDst, src):

        orig_is_target = self._is_target
        try:
            self._is_target = True
            dst_hdl = self.as_hdl(dst)
        finally:
            self._is_target = orig_is_target

        src_hdl = self.as_hdl_Value(src)
        if typeOfDst == SIGNAL_TYPE.REG:
            return HdlStmAssign(src_hdl, dst_hdl)
        else:
            return hdl_call(hdl_getattr(dst_hdl, "write"), [src_hdl, ])

    def as_hdl_HdlAssignmentContainer(self, a: HdlAssignmentContainer):
        dst = a.dst
        assert isinstance(dst, SignalItem)
        # assert not dst.virtual_only, "should not be required"

        if a.indexes is not None:
            for i in a.indexes:
                dst = dst[i]

        typeOfDst = systemCTypeOfSig(dst)
        if dst.virtual_only and isinstance(a.src, Operator):
            assert a.src.operator == AllOps.CONCAT
            return self._as_hdl_HdlAssignmentContainer(dst, typeOfDst, a.src.operands)

        if dst._dtype == a.src._dtype or (
                isinstance(dst._dtype, Bits) and a.src._dtype == BOOL):
            return self._as_hdl_HdlAssignmentContainer(dst, typeOfDst, a.src)
        else:
            raise SerializerException("%r <= %r  is not valid assignment\n"
                                      " because types are different (%r; %r) "
                                      % (dst, a.src, dst._dtype, a.src._dtype))



from collections import deque

from hwt.hdl.constants import NOP
from hwt.simulator.agentBase import SyncAgentBase
from hwtSimApi.hdlSimulator import HdlSimulator
from hwtSimApi.triggers import WaitCombRead, WaitWriteOnly, WaitCombStable
from pyMathBitPrecise.bit_utils import ValidityError


class VldSyncedAgent(SyncAgentBase):

    def __init__(self, sim: HdlSimulator, intf, allowNoReset=False):
        super(VldSyncedAgent, self).__init__(
            sim,
            intf,
            allowNoReset=allowNoReset)
        self.data = deque()
        self._vld = self.get_valid_signal(intf)

    def get_data(self):
        return self.intf.data.read()

    def set_data(self, data):
        self.intf.data.write(data)

    @classmethod
    def get_valid_signal(cls, intf):
        return intf.vld

    def get_valid(self):
        return self._vld.read()

    def set_valid(self, val):
        self._lastVld = val
        return self._vld.write(val)

    def setEnable_asDriver(self, en):
        super(VldSyncedAgent, self).setEnable_asDriver(en)
        if not en:
            self.set_valid(0)

    def monitor(self):
        yield WaitCombStable()
        if self.notReset():
            intf = self.intf
            vld = self.get_valid()
            try:
                vld = int(vld)
            except ValidityError:
                raise ValidityError(self.sim.now, self.intf._getFullName(),
                                    "vld signal in invalid state (would cause desynchronisation)")
            if vld:
                d = self.get_data()
                if self._debugOutput is not None:
                    self._debugOutput.write("%s, read, %d: %r\n" % (
                        intf._getFullName(),
                        self.sim.now, d))
                self.data.append(d)

    def driver(self):
        yield WaitCombRead()
        if self.data and self.notReset():
            d = self.data.popleft()
        else:
            d = NOP

        yield WaitWriteOnly()
        if d is NOP:
            self.set_data(None)
            self.set_valid(0)
        else:
            self.set_data(d)
            self.set_valid(1)
            if self._debugOutput is not None:
                self._debugOutput.write("%s, wrote, %d: %r\n" % (
                    self.intf._getFullName(),
                    self.sim.now, self.actualData))



from ipCorePackager.constants import DIRECTION
from hdlConvertorAst.hdlAst._expr import HdlDirection
from hwt.doc_markers import internal

HWT_TO_HDLCONVEROTR_DIRECTION = {
    DIRECTION.IN: HdlDirection.IN,
    DIRECTION.INOUT: HdlDirection.INOUT,
    DIRECTION.OUT: HdlDirection.OUT,
}


@internal
class TmpVarsSwap():
    """
    An object which is used as a context manager for tmpVars inside of :class:`~.ToHdlAst`
    """

    def __init__(self, ctx, tmpVars):
        self.ctx = ctx
        self.tmpVars = tmpVars

    def __enter__(self):
        self.orig = self.ctx.tmpVars
        self.ctx.tmpVars = self.tmpVars

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.ctx.tmpVars = self.orig



from typing import Tuple, List, Dict, Union, Optional

from hwt.doc_markers import internal
from hwt.hdl.operatorUtils import replace_input_in_expr
from hwt.hdl.sensitivityCtx import SensitivityCtx
from hwt.hdl.statements.statement import HdlStatement
from hwt.hdl.value import HValue
from hwt.hdl.valueUtils import isSameHVal, areSameHVals
from hwt.pyUtils.uniqList import UniqList
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


class HdlAssignmentContainer(HdlStatement):
    """
    Assignment container

    :ivar ~.src: source
    :ivar ~.dst: destination signal
    :ivar ~.indexes: description of index selector on dst
        (list of Index/Slice objects) (f.e. [0, 1] means  dst[0][1])

    :cvar __instCntr: counter used for generating instance ids
    :ivar ~._instId: internally used only for intuitive sorting of statements
    """

    _DEEPCOPY_SKIP = (*HdlStatement._DEEPCOPY_SKIP, 'src', 'dst', 'indexes')
    _DEEPCOPY_SHALLOW_ONLY = (*HdlStatement._DEEPCOPY_SHALLOW_ONLY, "indexes")

    __instCntr = 0

    def __init__(self, src: Union[RtlSignalBase, HValue], dst: RtlSignalBase,
                 indexes: Optional[List[Union[RtlSignalBase, HValue]]]=None, virtual_only=False,
                 parentStm: Optional[HdlStatement]=None,
                 sensitivity: Optional[UniqList]=None,
                 event_dependent_from_branch:Optional[int]=None):
        """
        :param dst: destination to assign to
        :param src: source which is assigned from
        :param indexes: description of index selector on dst
            (list of Index/Slice objects) (f.e. [[0], [1]] means  dst[0][1])
        :param virtual_only: flag indicates that this assignments
            is only virtual and should not be added into
            netlist, because it is only for internal notation
        """
        super(HdlAssignmentContainer, self).__init__(
            parentStm,
            sensitivity,
            event_dependent_from_branch=event_dependent_from_branch)
        self._instId = HdlAssignmentContainer._nextInstId()

        self.src = src
        self.dst = dst
        assert isinstance(dst, RtlSignalBase), dst

        isReal = not virtual_only

        if isinstance(src, RtlSignalBase):
            self._inputs.append(src)
            if isReal:
                src.endpoints.append(self)

        self.indexes = indexes
        if indexes:
            for i in indexes:
                if isinstance(i, RtlSignalBase):
                    self._inputs.append(i)
                    if isReal:
                        i.endpoints.append(self)

        if isReal:
            dst.drivers.append(self)
            dst.ctx.statements.add(self)

        self._outputs.append(dst)

    def __deepcopy__(self, memo: dict):
        result = super(HdlAssignmentContainer, self).__deepcopy__(memo)
        result.src = self.src
        result.dst = self.dst
        result._instId = self._nextInstId()
        return result

    @internal
    def _cut_off_drivers_of(self, sig: RtlSignalBase):
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._cut_off_drivers_of`
        """
        if self.dst is sig:
            self.parentStm = None
            return self
        else:
            return None

    @internal
    def _discover_enclosure(self) -> None:
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._discover_enclosure`
        """
        assert self._enclosed_for is None
        self._enclosed_for = set()
        self._enclosed_for.update(self._outputs)

    @internal
    def _discover_sensitivity(self, seen: set) -> None:
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._discover_sensitivity`
        """
        assert self._sensitivity is None
        ctx = self._sensitivity = SensitivityCtx()

        casualSensitivity = set()
        for inp in self._inputs:
            if inp not in seen:
                seen.add(inp)
                inp._walk_sensitivity(casualSensitivity, seen, ctx)
        ctx.extend(casualSensitivity)

    @internal
    def _fill_enclosure(self, enclosure: Dict[RtlSignalBase, HdlStatement]):
        """
        The assignment does not have any uncovered code branches

        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._fill_enclosure`
        """
        pass

    @internal
    def _iter_stms(self):
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._iter_stms`
        """
        return
        yield

    @internal
    def _on_parent_event_dependent(self):
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._on_parent_event_dependent`
        """
        self._event_dependent_from_branch = 0

    @internal
    def _try_reduce(self) -> Tuple[List[HdlStatement], bool]:
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._try_reduce`
        """
        return [self, ], False

    @internal
    def _is_mergable(self, other: HdlStatement) -> bool:
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._is_mergable`
        """
        return isinstance(other, self.__class__)

    def isSame(self, other):
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement.isSame`
        """
        if isinstance(other, self.__class__):
            if isSameHVal(self.dst, other.dst)\
                    and isSameHVal(self.src, other.src)\
                    and areSameHVals(self.indexes, other.indexes):
                return True
        return False

    @internal
    @classmethod
    def _nextInstId(cls):
        """
        Get next instance id
        """
        i = cls.__instCntr
        cls.__instCntr += 1
        return i

    @internal
    def _replace_input(self, toReplace: RtlSignalBase,
                       replacement: RtlSignalBase) -> None:
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._replace_input`
        """
        isTopStatement = self.parentStm is None

        if self.indexes:
            new_indexes = []
            for ind in self.indexes:
                new_i = replace_input_in_expr(self, ind, toReplace, replacement, isTopStatement)
                new_indexes.append(new_i)
            self.indexes = new_indexes

        self.src = replace_input_in_expr(self, self.src, toReplace, replacement, isTopStatement)

        self._replace_input_update_sensitivity_and_enclosure(toReplace, replacement)



from hwt.synthesizer.interfaceLevel.mainBases import InterfaceBase, UnitBase
from hwt.hdl.statements.statement import HdlStatement
from typing import TypeVar, Generic, Iterable, List, Union, Tuple

T = TypeVar("T", InterfaceBase, UnitBase, None)


class HObjList(list, Generic[T]):
    """
    Regular list with some interface/unit methods delegated on items.

    Main purpose of this class it let :class:`hwt.synthesizer.PropDeclrCollector.PropDeclrCollector`
    know that this is not an regular python array and that items should be registered as HW objects.

    :ivar _name: name of the property on parent
    :ivar _parent: parent Unit/Interface object
    :note: this object may be nested in HObjList instances but the parent and name will always corresponds
        to a Unit/Interface object, if there is any

    :note: :class:`hwt.synthesizer.PropDeclrCollector.PropDeclrCollector` is used by
        :class:`hwt.synthesizer.interface.Interface` and :class:`hwt.synthesizer.unit.Unit`
    """

    def __init__(self, *args, **kwargs):
        hdl_name_override = kwargs.pop("hdl_name", None)
        list.__init__(self, *args, **kwargs)
        self._name = None
        self._parent = None
        self._hdl_name_override = hdl_name_override

    def _on_append(self, self_obj: "HObjList", item: T, index: int):
        pass

    def _m(self) -> "HObjList":
        for item in self:
            item._m()
        return self

    def append(self, item: T):
        if self._on_append is not HObjList._on_append:
            self._on_append(self, item, len(self))
        return list.append(self, item)

    def clear(self, *args, **kwargs):
        assert self._parent is None
        return list.clear(self, *args, **kwargs)

    def extend(self, iterable: Iterable[T]):
        if self._on_append is not HObjList._on_append:
            offset = len(self)
            for i, item in enumerate(iterable):
                self._on_append(self, item, offset + i)
        return list.extend(self, iterable)

    def insert(self, *args, **kwargs):
        assert self._parent is None
        return list.insert(self, *args, **kwargs)

    def pop(self, *args, **kwargs) -> T:
        assert self._parent is None
        return list.pop(self, *args, **kwargs)

    def remove(self, *args, **kwargs):
        assert self._parent is None
        return list.remove(self, *args, **kwargs)

    def reverse(self, *args, **kwargs):
        assert self._parent is None
        return list.remove(self, *args, **kwargs)

    def sort(self, *args, **kwargs):
        assert self._parent is None
        return list.remove(self, *args, **kwargs)

    def _getHdlName(self):
        """Get name in HDL """

        # list of name or tulple (name, separator)
        name: List[Union[str, Tuple[str, str]]] = []
        tmp = self
        while isinstance(tmp, (InterfaceBase, HObjList)):
            n = tmp._name

            if name:
                name_sep = getattr(tmp, "_NAME_SEPARATOR", "_")
                n = (n, name_sep)
            else:
                # no need to add separator at the end because this is a last part of the name
                n = n

            add_name_part_from_this = True
            name_override = tmp._hdl_name_override
            if name_override is not None:
                # recursively apply renames
                if isinstance(name_override, str):
                    if isinstance(n, tuple) and name_override:
                        n = (name_override, n[1])
                    else:
                        n = name_override
                elif isinstance(name_override, dict):
                    last_name = name[-1]
                    if isinstance(last_name, tuple):
                        last_name = last_name[0]
                    no = name_override.get(last_name, None)
                    if no is not None:
                        if isinstance(no, str):
                            # everything what we resolved so far is overriden on this parent
                            name = [no, ]
                            add_name_part_from_this = False
                        else:
                            raise NotImplementedError()

                else:
                    raise TypeError(name_override)

            if add_name_part_from_this:
                name.append(n)
            tmp = getattr(tmp, "_parent", None)

        _name = []
        for n in reversed(name):
            if isinstance(n, str):
                _name.append(n)
            else:
                _name.extend(n)

        return "".join(_name)

    def _getFullName(self) -> str:
        """get all name hierarchy separated by '.' """
        name = ""
        tmp = self
        while isinstance(tmp, (InterfaceBase, HObjList)):
            n = tmp._name

            if name == '':
                name = n
            else:
                name = n + "." + name

            tmp = getattr(tmp, "_parent", None)

        return name

    def _make_association(self, *args, **kwargs):
        """
        Delegate _make_association on items

        :note: doc in :func:`~hwt.synthesizer.interfaceLevel.propDeclCollector._make_association`
        """
        for o in self:
            o._make_association(*args, **kwargs)
        return self

    def _updateParamsFrom(self, *args, **kwargs):
        """
        :note: doc in :func:`~hwt.synthesizer.interfaceLevel.propDeclCollector._updateParamsFrom`
        """
        for o in self:
            if isinstance(o, (InterfaceBase, UnitBase, HObjList)):
                o._updateParamsFrom(*args, **kwargs)
        return self

    def __call__(self, other: List[T], exclude=None, fit=False):
        """
        () operator behaving as assingment operator
        """
        if not isinstance(other, (list, tuple)):
            raise TypeError(other)
        if len(self) != len(other):
            raise ValueError("Different number of interfaces in list",
                             len(self), len(other))

        statements = []
        for a, b in zip(self, other):
            stms = a(b, exclude=exclude, fit=fit)
            if isinstance(stms, HdlStatement):
                statements.append(stms)
            else:
                statements += stms

        return statements



from typing import Set, List, Dict, Optional, Callable

from hwt.doc_markers import internal
from hwt.hdl.statements.statement import HdlStatement
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.hdl.statements.assignmentContainer import HdlAssignmentContainer


class HdlAssignmentContainer_constructor():

    def __init__(self, src, dst):
        self.dst = dst
        self.src = src

    def __call__(self) -> HdlAssignmentContainer:
        return HdlAssignmentContainer(self.src, self.dst)


@internal
def fill_stm_list_with_enclosure(parentStm: Optional[HdlStatement],
                                 current_enclosure: Set[RtlSignalBase],
                                 statements: List[HdlStatement],
                                 do_enclose_for: List[RtlSignalBase],
                                 enclosure: Dict[RtlSignalBase, Callable[[], HdlStatement]])\
        ->List[HdlStatement]:
    """
    Apply enclosure on list of statements
    (fill all unused code branches with assignments from value specified by enclosure)

    :param parentStm: optional parent statement where this list is some branch
    :param current_enclosure: list of signals for which this statement list is enclosed
    :param statements: list of statements
    :param do_enclose_for: selected signals for which enclosure should be used
    :param enclosure: enclosure values for signals

    :attention: original statements parameter can be modified
    :return: new statements
    """
    if statements is None:
        statements = []

    for e_sig in do_enclose_for:
        if e_sig in current_enclosure:
            continue
        enclosed = False
        for stm in statements:
            if e_sig in stm._outputs:
                if e_sig not in stm._enclosed_for:
                    stm._fill_enclosure(enclosure)
                enclosed = True
                break
        # any statement was not related with this signal,
        if not enclosed:
            e = enclosure[e_sig]
            a: HdlStatement = e()
            assert isinstance(a, HdlStatement), a
            statements.append(a)

            if parentStm is not None:
                a._set_parent_stm(parentStm)

    return statements



from hwt.doc_markers import internal
from hwt.pyUtils.uniqList import UniqList


class SensitivityCtx(UniqList):
    """
    Sensitivity list used for resolution of sensitivity for statement instances

    :ivar ~.contains_ev_dependency: True if this contains event dependent
        sensitivity
    """

    def __init__(self, initSeq=None):
        UniqList.__init__(self, initSeq=initSeq)
        self.contains_ev_dependency = False

    @internal
    def extend(self, items):
        UniqList.extend(self, items)
        if isinstance(items, SensitivityCtx):
            self.contains_ev_dependency |= items.contains_ev_dependency

    @internal
    def clear(self):
        UniqList.clear(self)
        self.contains_ev_dependency = False



from typing import Union, Tuple, Optional, List

from hwt.doc_markers import internal
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.hdlType import HdlType
from hwt.hdl.types.struct import HStruct, HStructFieldMeta, HStructField
from hwt.interfaces.std import BramPort_withoutClk, RegCntrl, VldSynced, Signal
from hwt.interfaces.structIntf import StructIntf
from hwt.interfaces.unionIntf import UnionSink, UnionSource
from hwt.synthesizer.hObjList import HObjList
from hwt.synthesizer.interface import Interface
from hwt.synthesizer.interfaceLevel.mainBases import InterfaceBase
from hwt.synthesizer.interfaceLevel.unitImplHelpers import getSignalName
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.synthesizer.typePath import TypePath


class IntfMap(list):
    """
    Container of interface map. The interface map object describes
    the data type using existing interface instances.

    Items can be Interface/RtlSignal or (type/interface/None/IntfMap, name).
    None is used for padding.
    """
    pass


@internal
def _HTypeFromIntfMap(intf: Union[RtlSignalBase, VldSynced, RegCntrl, BramPort_withoutClk, StructIntf, UnionSink, UnionSource]):
    name = getSignalName(intf)
    if isinstance(intf, (RtlSignalBase, Signal)):
        dtype = intf._dtype
    elif isinstance(intf, VldSynced):
        dtype = intf.data._dtype
    elif isinstance(intf, RegCntrl):
        dtype = intf.din._dtype
    elif isinstance(intf, BramPort_withoutClk):
        dtype = Bits(int(intf.DATA_WIDTH))[2 ** int(intf.ADDR_WIDTH)]
    else:
        dtype, name = intf
        assert isinstance(dtype, HdlType)
        assert isinstance(name, str)

    return (dtype, name)


@internal
def HTypeFromIntfMapItem(
        interfaceMapItem: Union[Tuple[HdlType, Optional[str]],
                                Tuple[Union[Interface, RtlSignalBase], str],
                                Tuple[List[Union[Interface, RtlSignalBase]], str],
                                Union[Interface, RtlSignalBase]]):
    isTerminal = False
    if isinstance(interfaceMapItem, (InterfaceBase, RtlSignalBase)):
        dtype, nameOrPrefix = _HTypeFromIntfMap(interfaceMapItem)
        isTerminal = True
    else:
        typeOrListOfInterfaces, nameOrPrefix = interfaceMapItem

        if isinstance(typeOrListOfInterfaces, list) and \
                not isinstance(typeOrListOfInterfaces, IntfMap):
            # list of HType instances for array
            parts = []
            arrayItem_t = None

            for item in typeOrListOfInterfaces:
                if isinstance(item, IntfMap):
                    t = HTypeFromIntfMap(item)
                else:
                    t = HTypeFromIntfMapItem(item).dtype
                if arrayItem_t is None:
                    arrayItem_t = t
                else:
                    assert arrayItem_t == t, (
                        "all items in array has to have same type", arrayItem_t, t)
                parts.append(t)

            dtype = arrayItem_t[len(parts)]

        elif isinstance(typeOrListOfInterfaces, HdlType):
            dtype = typeOrListOfInterfaces
            isTerminal = True
        elif isinstance(typeOrListOfInterfaces,
                        (InterfaceBase, RtlSignalBase)):
            # renamed interface, ignore original name
            dtype = _HTypeFromIntfMap(typeOrListOfInterfaces)[0]
            isTerminal = True
        elif isinstance(typeOrListOfInterfaces, IntfMap):
            dtype = HTypeFromIntfMap(typeOrListOfInterfaces)
        else:
            # tuple (tuple of interfaces, prefix)
            assert isinstance(typeOrListOfInterfaces,
                              tuple), typeOrListOfInterfaces
            dtype = HTypeFromIntfMap(typeOrListOfInterfaces)

    assert isinstance(nameOrPrefix, str) or nameOrPrefix is None, nameOrPrefix
    assert isinstance(dtype, HdlType)

    f = HStructField(dtype, nameOrPrefix)

    if not isTerminal:
        f.meta = HStructFieldMeta(split=True)

    return f


def HTypeFromIntfMap(interfaceMap: IntfMap) -> HStruct:
    """
    Generate flattened register map for HStruct

    :param interfaceMap: sequence of
        a tuple (HdlType, name) (will create HStructField) or
        a tuple (HdlType, None) (will create HStructField as padding) or
        a tuple (list of Interface instances, name) (will create HStructField of HStruct type) or
        an Interface instance (will create a HStructField for an interface, with a name of interface)
    :param DATA_WIDTH: width of word
    :param terminalNodes: None or set whre are placed StructField instances
        which are derived directly from interface
    :return: generator of tuple (type, name, BusFieldInfo)
    """
    structFields = []

    for m in interfaceMap:
        f = HTypeFromIntfMapItem(m)
        structFields.append(f)

    return HStruct(*structFields)


def isPaddingInIntfMap(item):
    if isinstance(item, HdlType):
        return True
    else:
        try:
            if isinstance(item, tuple):
                _item, name = item
                if name is None:
                    return True
        except ValueError:
            pass

    return False


def _walkStructIntfAndIntfMap_unpack(structIntf: Union[HObjList, StructIntf, UnionSink, UnionSource], intfMap):
    """
    Try to unpack intfMap and apply the selection on structIntf

    :return: Optional tuple Interface, intfMap
    """
    # items are Interface/RtlSignal or (type/interface/None or list of items, name)
    if isPaddingInIntfMap(intfMap):
        return
    elif isinstance(intfMap, tuple):
        item, name = intfMap
    else:
        item = intfMap
        assert isinstance(item, (InterfaceBase, RtlSignalBase)), item
        name = getSignalName(item)

    if isinstance(item, HdlType):
        # this part of structIntf was generated from type descriptin
        # and we are re searching only for those parts which were generated
        # from Interface/RtlSignal
        return

    return getattr(structIntf, name), item


def walkStructIntfAndIntfMap(structIntf: Union[HObjList, StructIntf, UnionSink, UnionSource], intfMap):
    """
    Walk StructInterface and interface map
    and yield tuples (Interface in StructInterface, interface in intfMap)
    which are on same place

    :param structIntf: an interface to walk
    :param intfMap: interface map

    :note: typical usecase is when there is StructIntf generated from description in intfMap
        and then you need to connect interface from intfMap to structIntf
        and there you can use this function to iterate over interfaces which belongs together
    """

    if isinstance(intfMap, (InterfaceBase, RtlSignalBase)):
        yield structIntf, intfMap
        return
    elif isinstance(intfMap, tuple):
        item = _walkStructIntfAndIntfMap_unpack(structIntf, intfMap)
        if item is None:
            # is padding or there is no interface specified for it in intfMap
            return
        else:
            structIntf, item = item
            yield from walkStructIntfAndIntfMap(structIntf, item)
    elif isinstance(structIntf, list):
        structIntf
        assert len(structIntf) == len(intfMap)
        for sItem, item in zip(structIntf, intfMap):
            yield from walkStructIntfAndIntfMap(sItem, item)
    else:
        assert isinstance(intfMap, IntfMap), intfMap
        for item in intfMap:
            _item = _walkStructIntfAndIntfMap_unpack(structIntf, item)
            if _item is not None:
                sItem, _item = _item
                yield from walkStructIntfAndIntfMap(sItem, _item)

def IntfMapItem_find_by_name(intf_map_item, name):
    if isinstance(intf_map_item, HStructField):
        intf_map_item = intf_map_item.dtype
    if isinstance(intf_map_item, IntfMap):
        for x in intf_map_item:
            if isinstance(x, InterfaceBase):
                if x._name == name:
                    return x
            elif isinstance(x, RtlSignalBase):
                if x.name == name:
                    return x
            else:
                v, n = x
                if n == name:
                    return v
        raise KeyError(name)
    elif isinstance(intf_map_item, HStruct):
        for f in intf_map_item.fields:
            if f.name == name:
                return f
        raise KeyError(name)
    else:
        raise NotImplementedError(intf_map_item)

def IntfMap_get_by_field_path(root: IntfMap, field_path: TypePath):
    actual = root
    # find in interfaceMap, skip first because it is the type itself
    for rec in field_path:
        #if isinstance(rec, (InterfaceBase, RtlSignalBase)):
        #    shouldEnter, shouldUse = False, True
        #    return shouldEnter, shouldUse
        if isinstance(rec, int):
            actual = actual[rec]
        elif isinstance(rec, str):
            actual = IntfMapItem_find_by_name(actual, rec)
        else:
            raise NotImplementedError(rec)
    return  actual




from hwt.doc_markers import internal
from hwt.hdl.operator import Operator
from hwt.hdl.operatorDefs import isEventDependentOp
from hwt.hdl.sensitivityCtx import SensitivityCtx
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


@internal
def discoverEventDependency(sig: RtlSignalBase):
    """
    :return: generator of tuples (event operator, signal)
    """

    try:
        drivers = sig.drivers
    except AttributeError:
        return

    if len(drivers) == 1:
        d = drivers[0]
        if isinstance(d, Operator):
            if isEventDependentOp(d.operator):
                yield (d.operator, d.operands[0])
            else:
                for op in d.operands:
                    yield from discoverEventDependency(op)


@internal
def discover_sensitivity_of_sig(signal: RtlSignalBase,
                              seen: set, ctx: SensitivityCtx):
    casualSensitivity = set()
    signal._walk_sensitivity(casualSensitivity, seen, ctx)
    if not ctx.contains_ev_dependency:
        # if event dependent sensitivity found do not add other sensitivity
        ctx.extend(casualSensitivity)



from copy import deepcopy, copy
from functools import reduce
from itertools import compress
from operator import and_
from typing import List, Tuple, Dict, Optional, Callable, Set

from hwt.doc_markers import internal
from hwt.hdl.operatorUtils import replace_input_in_expr
from hwt.hdl.sensitivityCtx import SensitivityCtx
from hwt.hdl.statements.statement import HdlStatement, HwtSyntaxError
from hwt.hdl.statements.utils.comparison import isSameStatementList, statementsAreSame
from hwt.hdl.statements.utils.ioDiscovery import HdlStatement_discover_enclosure_for_statements
from hwt.hdl.statements.utils.reduction import HdlStatement_merge_statement_lists, \
    HdlStatement_try_reduce_list, is_mergable_statement_list
from hwt.hdl.statements.utils.signalCut import HdlStatement_cut_off_drivers_of_list
from hwt.hdl.types.enum import HEnum
from hwt.hdl.value import HValue
from hwt.hdl.valueUtils import isSameHVal
from hwt.serializer.utils import RtlSignal_sort_key
from hwt.synthesizer.rtlLevel.fill_stm_list_with_enclosure import fill_stm_list_with_enclosure
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.synthesizer.rtlLevel.rtlSignal import RtlSignal


class SwitchContainer(HdlStatement):
    """
    Structural container for switch statement for hdl rendering

    :ivar ~.switchOn: select signal of switch
    :ivar ~.cases: list of tuples (value, statements)
    :ivar ~.default: list of statements (for branch "default")
    :ivar ~._case_value_index: dictionary {value:index} for every case in cases
    :ivar ~._case_enclosed_for: list of sets of enclosed signal for each case branch
    :ivar ~._default_enclosed_for: set of enclosed signals for branch default
    """
    _DEEPCOPY_SKIP = (*HdlStatement._DEEPCOPY_SKIP, 'switchOn', 'cases')
    _DEEPCOPY_SHALLOW_ONLY = (*HdlStatement._DEEPCOPY_SHALLOW_ONLY, '_case_value_index', '_case_enclosed_for', '_default_enclosed_for')

    def __init__(self, switchOn: RtlSignal,
                 cases: List[Tuple[HValue, List[HdlStatement]]],
                 default: List[HdlStatement]=None,
                 parentStm: HdlStatement=None,
                 event_dependent_from_branch: Optional[int]=None):

        super(SwitchContainer, self).__init__(
            parentStm=parentStm,
            event_dependent_from_branch=event_dependent_from_branch)
        self.switchOn = switchOn
        self.cases = cases
        self.default = default

        self._case_value_index = {}
        for i, (v, _) in enumerate(cases):
            assert v not in self._case_value_index, v
            self._case_value_index[v] = i

        self._case_enclosed_for: Optional[List[Set[RtlSignal]]] = None
        self._default_enclosed_for: Optional[Set[RtlSignal]] = None

    def __deepcopy__(self, memo: dict):
        result = super(SwitchContainer, self).__deepcopy__(memo)
        result.switchOn = self.switchOn
        result.cases = [(c, deepcopy(stms, memo)) for c, stms in self.cases]
        result._case_value_index = copy(self._case_value_index)
        result._case_enclosed_for = copy(self._case_enclosed_for)
        result._default_enclosed_for = copy(self._default_enclosed_for)
        return result

    @internal
    def _cut_off_drivers_of(self, sig: RtlSignalBase):
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._cut_off_drivers_of`
        """
        if self._sensitivity is not None or self._enclosed_for is not None:
                raise NotImplementedError(
                    "Sensitivity and enclosure has to be cleaned first")

        if len(self._outputs) == 1 and sig in self._outputs:
            # this statement has only this output, eject this statement from its parent
            self.parentStm = None  # because new parent will be asigned immediately after cutting of
            return self

        sig.drivers.discard(self)
        # try to cut off all statements which are drivers of specified signal
        # in all branches
        child_keep_mask = []

        all_cut_off = True
        new_default = None
        if self.default:
            new_default = []
            child_keep_mask.clear()
            case_eliminated = HdlStatement_cut_off_drivers_of_list(
                sig, self.default, child_keep_mask, new_default)
            all_cut_off &= case_eliminated
            if case_eliminated:
                self.rank -= 1
                self.default = None
            else:
                self.default = list(compress(self.default, child_keep_mask))

        new_cases = []
        case_keepmask = []
        for val, stms in self.cases:
            new_case = []
            child_keep_mask.clear()
            case_eliminated = HdlStatement_cut_off_drivers_of_list(
                sig, stms, child_keep_mask, new_case)
            if case_eliminated:
                self.rank -= 1

            all_cut_off &= case_eliminated
            case_keepmask.append(not case_eliminated)

            _stms = list(compress(stms, child_keep_mask))
            stms.clear()
            stms.extend(_stms)

            if new_case or new_default:
                # if there is a default we need to add case even in empty
                # to prevent falling to default
                new_cases.append((val, new_case))

        self.cases = list(compress(self.cases, case_keepmask))

        assert not all_cut_off, "everything was cut of but this should be already known at start"

        if new_cases or new_default:
            # parts were cut off
            # generate new statement for them
            sel_sig = self.switchOn
            n = self.__class__(sel_sig)
            n.add_cases(new_cases)
            if new_default:
                n.Default(*new_default)

            if self.parentStm is None:
                ctx = n._get_rtl_context()
                ctx.statements.add(n)

            self._cut_off_drivers_of_regenerate_io(sig, n)

            return n

    @internal
    def _clean_signal_meta(self):
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._clean_signal_meta`
        """
        self._case_enclosed_for = None
        self._default_enclosed_for = None
        HdlStatement._clean_signal_meta(self)

    @internal
    def _collect_io(self):
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._collect_io`
        """
        if isinstance(self.switchOn, RtlSignalBase):
            self._inputs.append(self.switchOn)
        for c, _ in self.cases:
            if isinstance(c, RtlSignalBase):
                self._inputs.append(c)
        super(SwitchContainer, self)._collect_io()

    @internal
    def _collect_inputs(self) -> None:
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._collect_inputs`
        """
        if isinstance(self.switchOn, RtlSignalBase):
            self._inputs.append(self.switchOn)
        for c, _ in self.cases:
            if isinstance(c, RtlSignalBase):
                self._inputs.append(c)
        super(SwitchContainer, self)._collect_inputs()

    @internal
    def _discover_enclosure(self) -> None:
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._discover_enclosure`
        """
        assert self._enclosed_for is None
        enclosure = self._enclosed_for = set()
        case_enclosures = self._case_enclosed_for = []
        outputs = self._outputs

        for _, stms in self.cases:
            c_e = HdlStatement_discover_enclosure_for_statements(stms, outputs)
            case_enclosures.append(c_e)

        self._default_enclosed_for = HdlStatement_discover_enclosure_for_statements(
            self.default, outputs)

        t = self.switchOn._dtype
        if not self.default and len(self.cases) < t.domain_size():
            # cases does not cover all branches
            return

        for s in outputs:
            enclosed = True
            for e in case_enclosures:
                if s not in e:
                    enclosed = False
                    break

            if enclosed and (not self.default or s in self._default_enclosed_for):
                enclosure.add(s)

    @internal
    def _discover_sensitivity(self, seen) -> None:
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._discover_sensitivity`
        """
        assert self._sensitivity is None, self
        ctx = self._sensitivity = SensitivityCtx()

        casual_sensitivity = set()
        self.switchOn._walk_sensitivity(casual_sensitivity, seen, ctx)
        if ctx.contains_ev_dependency:
            raise HwtSyntaxError(
                "Can not switch on event operator result", self.switchOn)
        ctx.extend(casual_sensitivity)

        for stm in self._iter_stms():
            stm._discover_sensitivity(seen)
            ctx.extend(stm._sensitivity)

    @internal
    def _fill_enclosure(self, enclosure: Dict[RtlSignalBase, Callable[[], HdlStatement]]) -> None:
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._fill_enclosure`
        """
        select = []
        outputs = self._outputs
        for e in sorted(enclosure.keys(), key=RtlSignal_sort_key):
            if e in outputs:
                select.append(e)

        for (_, stms), e in zip(self.cases, self._case_enclosed_for):
            fill_stm_list_with_enclosure(self, e, stms, select, enclosure)
            e.update(select)

        t = self.switchOn._dtype
        default_required = len(self.cases) < t.domain_size()

        if self.default is not None or default_required:
            self.default = fill_stm_list_with_enclosure(
                self, self._default_enclosed_for, self.default, select, enclosure)
            self._default_enclosed_for.update(select)

        self._enclosed_for.update(select)

    def _iter_stms(self):
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._iter_stms`
        """
        for _, stms in self.cases:
            yield from stms

        if self.default is not None:
            yield from self.default

    @internal
    def _is_mergable(self, other) -> bool:
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._is_mergable`
        """
        if not isinstance(other, SwitchContainer):
            return False

        if not (self.switchOn is other.switchOn and
                len(self.cases) == len(other.cases) and
                is_mergable_statement_list(self.default, other.default)):
            return False

        for (vA, caseA), (vB, caseB) in zip(self.cases, other.cases):
            if vA != vB or not is_mergable_statement_list(caseA, caseB):
                return False

        return True

    @internal
    def _merge_with_other_stm(self, other: "SwitchContainer") -> None:
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._merge_with_other_stm`
        """
        merge = HdlStatement_merge_statement_lists
        newCases = []
        for (c, caseA), (_, caseB) in zip(self.cases, other.cases):
            newCases.append((c, merge(caseA, caseB)))

        self.cases = newCases

        if self.default is not None:
            self.default = merge(self.default, other.default)

        self._on_merge(other)

    @internal
    def _try_reduce(self) -> Tuple[List["HdlStatement"], bool]:
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._try_reduce`
        """
        io_change = False

        # try reduce the content of the case branches
        new_cases = []
        for val, statements in self.cases:
            _statements, rank_decrease, _io_change = HdlStatement_try_reduce_list(
                statements)
            io_change |= _io_change
            self.rank -= rank_decrease
            new_cases.append((val, _statements))
        self.cases = new_cases

        # try reduce content of the defult branch
        if self.default is not None:
            self.default, rank_decrease, _io_change = HdlStatement_try_reduce_list(
                self.default)
            self.rank -= rank_decrease
            io_change |= _io_change

        # try reduce self
        reduce_self = not self._condHasEffect()
        if reduce_self:
            if self.cases:
                res = self.cases[0][1]
            elif self.default is not None:
                res = self.default
            else:
                res = []
        else:
            res = [self, ]

        self._on_reduce(reduce_self, io_change, res)

        if not self.default:
            t = self.switchOn._dtype
            if isinstance(t, HEnum):
                dom_size = t.domain_size()
                val_cnt = len(t._allValues)
                if len(self.cases) == val_cnt and val_cnt < dom_size:
                    # bit representation is not fully matching enum description
                    # need to set last case as default to prevent latches
                    _, stms = self.cases.pop()
                    self.default = stms

        return res, io_change

    @internal
    def _condHasEffect(self) -> bool:
        """
        :return: True if statements in branches has different effect
        """
        if not self.cases:
            return False

        # [TODO]
        type_domain_covered = bool(self.default) or len(
            self.cases) == self.switchOn._dtype.domain_size()

        stmCnt = len(self.cases[0][1])
        if type_domain_covered and reduce(
                and_,
                [len(stm) == stmCnt
                 for _, stm in self.cases],
                True) and (self.default is None
                           or len(self.default) == stmCnt):
            stms = list(self._iter_stms())
            if statementsAreSame(stms):
                return False
            else:
                return True
        return True

    @internal
    def _replace_input(self, toReplace: RtlSignalBase,
                       replacement: RtlSignalBase) -> None:
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._replace_input`
        """
        isTopStatement = self.parentStm is None
        self.switchOn = replace_input_in_expr(self, self.switchOn, toReplace,
                                              replacement, isTopStatement)

        for (_, stms) in self.cases:
            for stm in stms:
                stm._replace_input(toReplace, replacement)

        if self.default is not None:
            for stm in self.default:
                stm._replace_input(toReplace, replacement)

        self._replace_input_update_sensitivity_and_enclosure(toReplace, replacement)

    @internal
    def _replace_child_statement(self, stm: HdlStatement,
            replacement:List[HdlStatement],
            update_io:bool) -> None:
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._replace_child_statement`
        """

        if update_io:
            raise NotImplementedError()
        for branch_list in (*(case_stms for _, case_stms in self.cases), self.default):
            if branch_list is None:
                continue
            try:
                i = branch_list.index(stm)
            except ValueError:
                # not in list
                continue

            self.rank -= stm.rank
            branch_list[i:i + 1] = replacement
            for rstm in replacement:
                rstm._set_parent_stm(self)
            # reset IO because it was shared with this statement
            stm._destroy()
            return

        raise ValueError("Statement", stm, "not found in ", self)

    def isSame(self, other: HdlStatement) -> bool:
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement.isSame`
        """
        if self is other:
            return True

        if self.rank != other.rank:
            return False

        if isinstance(other, SwitchContainer) \
                and isSameHVal(self.switchOn, other.switchOn)\
                and len(self.cases) == len(other.cases)\
                and isSameStatementList(self.default, other.default):
            for (ac, astm), (bc, bstm) in zip(self.cases, other.cases):
                if not isSameHVal(ac, bc)\
                        or not isSameStatementList(astm, bstm):
                    return False
            return True
        return False



from hwt.hdl.types.bits import Bits
from hwt.interfaces.std import Signal
from hwt.synthesizer.interface import Interface


class DifferentialSig(Interface):
    """
    Interface of differential pair
    """

    def _declr(self):
        self.n = Signal(dtype=Bits(1, negated=True))
        self.p = Signal()



from itertools import chain
from typing import List, Tuple, Union, Optional, Dict, Callable

from hwt.doc_markers import internal
from hwt.hdl.hdlObject import HdlObject
from hwt.pyUtils.arrayQuery import flatten
from hwt.pyUtils.uniqList import UniqList
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from copy import deepcopy, copy


class HwtSyntaxError(Exception):
    pass


class HdlStatement(HdlObject):
    """
    :ivar ~._event_dependent_from_branch: index of code branch if statement is event (clk) dependent else None
    :ivar ~.parentStm: parent instance of HdlStatement or None
    :ivar ~._inputs: UniqList of input signals for this statement
    :ivar ~._outputs: UniqList of output signals for this statement
    :ivar ~._sensitivity: UniqList of input signals
        or (rising/falling) operator
    :ivar ~._enclosed_for: set of outputs for which this statement is enclosed
        (for which there is not any unused branch)
    :ivar ~.rank: number of used branches in statement, used as pre-filter
        for statement comparing
    """
    _DEEPCOPY_SKIP = ('parentStm',)
    _DEEPCOPY_SHALLOW_ONLY = ("_inputs", "_outputs", "_enclosed_for", "_sensitivity")

    def __init__(self, parentStm:Optional["HdlStatement"]=None,
                 sensitivity:Optional[UniqList]=None,
                 event_dependent_from_branch:Optional[int]=None):
        assert event_dependent_from_branch is None or isinstance(event_dependent_from_branch, int), event_dependent_from_branch
        self._event_dependent_from_branch = event_dependent_from_branch
        self.parentStm = parentStm
        self._inputs = UniqList()
        self._outputs = UniqList()
        self._enclosed_for = None

        self._sensitivity = sensitivity
        self.rank = 0

    def __deepcopy__(self, memo: dict):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if v is None:
                new_v = None
            elif k in self._DEEPCOPY_SKIP:
                new_v = None
            elif k in self._DEEPCOPY_SHALLOW_ONLY:
                new_v = copy(v)
            else:
                new_v = deepcopy(v, memo)

            setattr(result, k, new_v)
        return result

    @internal
    def _clean_signal_meta(self):
        """
        Clean informations about enclosure for outputs and sensitivity
        of this statement
        """
        self._enclosed_for = None
        self._sensitivity = None
        for stm in self._iter_stms():
            stm._clean_signal_meta()

    @internal
    def _collect_io(self) -> None:
        """
        Collect inputs/outputs from all child statements
        to :py:attr:`~_input` / :py:attr:`_output` attribute on this object
        """
        in_add = self._inputs.extend
        out_add = self._outputs.extend

        for stm in self._iter_stms():
            in_add(stm._inputs)
            out_add(stm._outputs)

    @internal
    def _collect_inputs(self) -> None:
        """
        Collect inputs from all child statements
        to :py:attr:`~_input` attribute on this object
        """
        in_add = self._inputs.extend

        for stm in self._iter_stms():
            in_add(stm._inputs)

    @internal
    def _collect_outputs(self) -> None:
        """
        Collect inputs from all child statements
        to :py:attr:`_output` attribute on this object
        """

        out_add = self._outputs.extend

        for stm in self._iter_stms():
            out_add(stm._outputs)

    @internal
    def _cut_off_drivers_of(self, sig: RtlSignalBase) -> Union[None, "HdlStatement", List["HdlStatement"]]:
        """
        Cut all logic from statements which drives signal sig.

        :param sig: signal which drivers should be removed
        :return: A statement or statement list which was cut off from the original statement
        """
        raise NotImplementedError("This is an abstract method and it should be implemented in child class",
                                  self.__class__)

    @internal
    def _cut_off_drivers_of_regenerate_io(self, cut_off_sig: RtlSignalBase, cut_of_smt: "HdlStatement"):
        """
        Update _inputs/_outputs after some part of statement was cut of

        :param cut_off_sig: a signal which driver is a cut_of_stm
        :param cut_of_smt: the statement wich was cut off from original statement (selected by cut_off_sig)
        """
        # update io of this
        self._outputs.remove(cut_off_sig)
        if cut_of_smt._inputs:
            # update inputs on this
            self._inputs.clear()
            self._collect_inputs()
            if self.parentStm is None:
                for i in cut_of_smt._inputs:
                    if i not in self._inputs:
                        i.endpoints.remove(self)

        if self.parentStm is None:
            cut_off_sig.drivers.append(cut_of_smt)

    @internal
    def _discover_enclosure(self) -> None:
        """
        Discover all outputs for which is this statement enclosed _enclosed_for property
        (has driver in all code branches)
        """
        raise NotImplementedError("This method should be implemented"
                                  " on class of statement", self.__class__, self)

    @internal
    def _discover_sensitivity(self, seen: set) -> None:
        """
        discover all sensitivity signals and store them to _sensitivity property
        """
        raise NotImplementedError("This method should be implemented"
                                  " on class of statement", self.__class__, self)

    @internal
    def _fill_enclosure(self, enclosure: Dict[RtlSignalBase, Callable[[], 'HdlStatement']]) -> None:
        """
        Add assignments to a default values to a code branches which are not assigning to specified output signal.

        :attention: enclosure has to be discoverd first use _discover_enclosure()  method
        """
        raise NotImplementedError("This method should be implemented"
                                  " on class of statement", self.__class__, self)

    @internal
    def _get_rtl_context(self) -> 'RtlNetlist':
        """
        get RtlNetlist context from signals
        """
        for sig in chain(self._inputs, self._outputs):
            if sig.ctx:
                return sig.ctx
            else:
                # Param instances does not have context
                continue
        raise HwtSyntaxError(
            "Statement does not have any signal in any context,"
            " it should have at least some output or should be alreary optimized out", self)

    def _iter_stms(self):
        """
        :return: iterator over all children statements
        """
        raise NotImplementedError("This method should be implemented"
                                  " on class of statement", self.__class__,
                                  self)

    @internal
    def _merge_with_other_stm(self, other: "HdlStatement") -> None:
        """
        :attention: statements has to be mergable (to check use _is_mergable method)
        """
        raise NotImplementedError("This method should be implemented"
                                  " on class of statement", self.__class__,
                                  self)

    @internal
    def _on_reduce(self, self_reduced: bool, io_changed: bool,
                   result_statements: List["HdlStatement"]) -> None:
        """
        Update signal IO after reduce attempt

        :param self_reduced: if True this object was reduced
        :param io_changed: if True IO of this object may changed
            and has to be updated
        :param result_statements: list of statements which are result
            of reduce operation on this statement
        """

        parentStm = self.parentStm
        if self_reduced:
            was_top = parentStm is None
            # update signal drivers/endpoints
            if was_top:
                # disconnect self from signals
                ctx = self._get_rtl_context()
                ctx.statements.remove(self)
                ctx.statements.update(result_statements)

                for i in self._inputs:
                    i.endpoints.discard(self)
                for o in self._outputs:
                    o.drivers.remove(self)

            for stm in result_statements:
                stm.parentStm = parentStm
                if parentStm is None:
                    # connect signals to child statements
                    for inp in stm._inputs:
                        inp.endpoints.append(stm)
                    for outp in stm._outputs:
                        outp.drivers.append(stm)
        else:
            # parent has to update it's inputs/outputs
            if io_changed:
                self._inputs = UniqList()
                self._outputs = UniqList()
                self._collect_io()

    @internal
    def _on_merge(self, other):
        """
        After merging statements update IO, sensitivity and context

        :attention: rank is not updated
        """
        self._inputs.extend(other._inputs)
        self._outputs.extend(other._outputs)

        if self._sensitivity is not None:
            self._sensitivity.extend(other._sensitivity)
        else:
            assert other._sensitivity is None

        if self._enclosed_for is not None:
            self._enclosed_for.update(other._enclosed_for)
        else:
            assert other._enclosed_for is None

        other_was_top = other.parentStm is None
        if other_was_top:
            other._get_rtl_context().statements.remove(other)
            for s in other._inputs:
                s.endpoints.discard(other)
                s.endpoints.append(self)

            for s in other._outputs:
                s.drivers.discard(other)
                s.drivers.append(self)

    @internal
    def _try_reduce(self) -> Tuple[List["HdlStatement"], bool]:
        raise NotImplementedError("This method should be implemented"
                                  " on class of statement", self.__class__,
                                  self)

    def _is_enclosed(self) -> bool:
        """
        :return: True if every branch in statement assignas to all output signals else False
        """
        return len(self._outputs) == len(self._enclosed_for)

    @internal
    def _is_mergable(self, other: "HdlStatement") -> bool:
        if self is other:
            raise ValueError("Can not merge statement with itself")
        else:
            raise NotImplementedError("This method should be implemented"
                                      " on class of statement", self.__class__,
                                      self)

    @internal
    def _on_parent_event_dependent(self):
        """
        After parent statement become event dependent
        propagate event dependency flag to child statements
        """
        if self._event_dependent_from_branch != 0:
            self._event_dependent_from_branch = 0
            for stm in self._iter_stms():
                stm._on_parent_event_dependent()

    @internal
    def _set_parent_stm(self, parentStm: "HdlStatement"):
        """
        Assign parent statement and propagate dependency flags if necessary
        """
        was_top = self.parentStm is None
        self.parentStm = parentStm
        if self._event_dependent_from_branch is None\
                and parentStm._event_dependent_from_branch is not None:
            self._on_parent_event_dependent()

        topStatement = parentStm
        parents = []
        while True:
            parents.append(topStatement)
            if topStatement.parentStm is None:
                break
            topStatement = topStatement.parentStm

        if was_top:
            for inp in self._inputs:
                inp.endpoints.discard(self)
                inp.endpoints.append(topStatement)
                for p in parents:
                    p._inputs.append(inp)

            for outp in self._outputs:
                outp.drivers.discard(self)
                outp.drivers.append(topStatement)
                for p in parents:
                    p._outputs.append(outp)

            ctx = self._get_rtl_context()
            ctx.statements.discard(self)

        parentStm.rank += self.rank

    @internal
    def _register_stements(self, statements: List["HdlStatement"],
                           target: List["HdlStatement"]):
        """
        Append statements to this container
        """
        for stm in flatten(statements):
            assert stm.parentStm is None, (
                "HdlStatement instance has to have only a single parent", stm)
            stm._set_parent_stm(self)
            target.append(stm)

    def isSame(self, other: "HdlStatement") -> bool:
        """
        :return: True if other has same meaning as self
        """
        raise NotImplementedError("This method should be implemented in child class", self.__class__, self)

    @internal
    def _destroy(self):
        """
        Disconnect this statement from signals and delete it from RtlNetlist context

        :attention: signal endpoints/drivers will be altered
            that means they can not be used for iteration
        """
        for i in self._inputs:
            i.endpoints.discard(self)

        if self.parentStm is None:
            ctx = self._get_rtl_context()
            for o in self._outputs:
                o.drivers.remove(self)

            ctx.statements.remove(self)
            self.parentStm = None

    @internal
    def _replace_input(self, toReplace: RtlSignalBase,
                       replacement: RtlSignalBase) -> None:
        """
        Replace input signal with another

        :note: sensitivity/endoints are actualized
        """
        raise NotImplementedError("This method should be implemented in child class", self.__class__, self)

    @internal
    def _replace_input_update_sensitivity_and_enclosure(
            self,
            toReplace: RtlSignalBase,
            replacement: RtlSignalBase):
        if self._sensitivity is not None:
            if self._sensitivity.discard(toReplace):
                self._sensitivity.add(replacement)

        if self._enclosed_for is not None:
            if self._enclosed_for.discard(toReplace):
                self._enclosed_for.add(replacement)

    @internal
    def _replace_child_statement(self, stm: "HdlStatement",
                                 replacement: List["HdlStatement"],
                                 update_io: bool) -> None:
        """
        Replace a child statement with a list of other statements

        :attention: original statement is destroyed and entirely removed from circuit
        :note: sensitivity/endoints are actualized
        """
        raise NotImplementedError("This method should be implemented in child class", self.__class__, self)




from hdlConvertorAst.hdlAst._expr import HdlOpType
from hwt.hdl.operatorDefs import AllOps


HWT_TO_HDLCONVEROTR_OPS = {
    **{op: getattr(HdlOpType, op.id) for op in [
        AllOps.AND,
        AllOps.OR,
        AllOps.XOR,
        AllOps.CONCAT,
        AllOps.DIV,
        AllOps.DOWNTO,
        AllOps.TO,
        AllOps.EQ,
        AllOps.GT,
        AllOps.GE,
        AllOps.LE,
        AllOps.POW,
        AllOps.LT,
        AllOps.SUB,
        AllOps.MUL,
        AllOps.NE,
        AllOps.ADD,
    ]},
    AllOps.NOT: HdlOpType.NEG,
    AllOps.MINUS_UNARY: HdlOpType.MINUS_UNARY,
    AllOps.RISING_EDGE: HdlOpType.RISING,
    AllOps.FALLING_EDGE: HdlOpType.FALLING,
}



from hwt.doc_markers import internal
from hwt.hdl.constants import INTF_DIRECTION
from hwt.synthesizer.unit import Unit
from hwtSimApi.hdlSimulator import HdlSimulator


@internal
def autoAddAgents(unit: Unit, sim: HdlSimulator):
    """
    Walk all interfaces on unit and instantiate agent for every interface.

    :return: all monitor/driver functions which should be added to simulation
         as processes
    """
    for intf in unit._interfaces:
        assert intf._isExtern, intf

        intf._initSimAgent(sim)
        assert intf._ag is not None, intf


@internal
def collect_processes_from_sim_agents(unit: Unit):
    proc = []
    for intf in unit._interfaces:
        a = intf._ag
        if not intf._isExtern or a is None:
            continue

        if intf._direction == INTF_DIRECTION.MASTER:
            agProcs = a.getMonitors()
        elif intf._direction == INTF_DIRECTION.SLAVE:
            agProcs = a.getDrivers()
        else:
            raise NotImplementedError(f"intf._direction {intf._direction} for {intf}")

        proc.extend(agProcs)

    return proc




"""
This package contains serializers. Purpose of serializer class
is to convert hwt representations of designed architecture
to hdlConvertorAst to convert it to target language (VHDL/Verilog/SystemC...+xdc/ucf/...).

Rather than using serializer classes manually it is recommended to use :func:`hwt.synthesizer.utils.to_rtl`

The serialization process is usually a destructive operation as parts of AST can be rewritten
to fit target language.
"""



from collections import deque
from inspect import isgenerator
import sys
from typing import Union

from hwt.doc_markers import internal
from hwt.hdl.types.arrayVal import HArrayVal
from hwt.hdl.value import HValue
from hwt.serializer.generic.indent import getIndent
from hwt.synthesizer.hObjList import HObjList
from hwt.synthesizer.interfaceLevel.mainBases import InterfaceBase
from hwt.synthesizer.unit import Unit
from pyMathBitPrecise.bits3t import Bits3val


def pprintInterface(intf: Union[Unit, InterfaceBase], indent:int=0, prefix:str="", file=sys.stdout):
    """
    Pretty print interface
    """
    try:
        s = intf._sig
    except AttributeError:
        s = None
    if s is None:
        s = ""
    else:
        s = " " + repr(s)

    file.write("".join([getIndent(indent), prefix, repr(intf._getFullName()),
                        s]))
    file.write("\n")

    if isinstance(intf, HObjList):
        for i, p in enumerate(intf):
            # interfaces have already name of this array and index in it's name
            pprintInterface(p, indent=indent + 1, prefix=prefix, file=file)
    else:
        for i in intf._interfaces:
            pprintInterface(i, indent=indent + 1, file=file)


def pprintAgents(unitOrIntf: Union[Unit, InterfaceBase], indent:int=0, prefix:str="", file=sys.stdout):
    if isinstance(unitOrIntf, InterfaceBase):
        ag = unitOrIntf._ag
    elif isinstance(unitOrIntf, HObjList):
        for i, item in enumerate(unitOrIntf):
            item_prefix = f"{prefix}_{i:d}"
            pprintAgents(item, indent=indent + 1, prefix=item_prefix, file=file)
        return
    else:
        ag = None

    if ag is not None:
        indent_str = getIndent(indent)
        file.write(f"{indent_str:s}{prefix:s}{ag}\n")

    for i in unitOrIntf._interfaces:
        pprintAgents(i, indent + 1, file=file)


@internal
def reconnectUnitSignalsToModel(synthesisedUnitOrIntf: Union[Unit, InterfaceBase], rtl_simulator):
    """
    Reconnect model signals to unit to run simulation with simulation model
    but use original unit interfaces for communication

    :param synthesisedUnitOrIntf: interface where should be signals
        replaced from signals from modelCls
    :param rtl_simulator: RTL simulator form where signals
        for synthesisedUnitOrIntf should be taken
    """
    obj = synthesisedUnitOrIntf

    for intf in obj._interfaces:
        if intf._interfaces:
            reconnectUnitSignalsToModel(intf, rtl_simulator)
        else:
            # reconnect signal from model
            name = intf._sigInside.name
            # update name and dtype
            s = getattr(rtl_simulator.io, name)
            if s._dtype is None:
                s._dtype = intf._dtype
            s._name = intf._name
            s.name = name
            intf.read = s.read
            intf.write = s.write
            intf.wait = s.wait
            intf._sigInside = s


def valuesToInts(values):
    """
    Iterable of values to ints (nonvalid = None)
    """
    return [valToInt(d) for d in values]


def valToInt(v):
    try:
        return int(v)
    except ValueError:
        return None

def allValuesToInts(sequenceOrVal):
    """
    Convert HValue instances to int recursively (for sequences)
    """
    if isinstance(sequenceOrVal, HArrayVal):
        sequenceOrVal = sequenceOrVal.val

    if isinstance(sequenceOrVal, (HValue, Bits3val)):
        return valToInt(sequenceOrVal)
    elif not sequenceOrVal:
        return sequenceOrVal
    elif (isinstance(sequenceOrVal, (list, tuple, deque))
          or isgenerator(sequenceOrVal)):
        seq = []
        for i in sequenceOrVal:
            seq.append(allValuesToInts(i))

        if isinstance(sequenceOrVal, tuple):
            return tuple(seq)

        return seq
    else:
        return sequenceOrVal



from ipCorePackager.constants import INTF_DIRECTION
from enum import Enum


class SignalDriverErrType(Enum):
    (
        MISSING_DRIVER,
        MULTIPLE_COMB_DRIVERS,
        OUTPUT_WITHOUT_DRIVER,
        INPUT_WITH_DRIVER
    ) = range(4)


SignalDriverErrType_labels = {
    SignalDriverErrType.MISSING_DRIVER: "  missing driver for:",
    SignalDriverErrType.MULTIPLE_COMB_DRIVERS: "  multiple comb. drivers for:",
    SignalDriverErrType.OUTPUT_WITHOUT_DRIVER: "  outputs without driver:",
    SignalDriverErrType.INPUT_WITH_DRIVER: "  inputs with driver:",
}


class SignalDriverErr(Exception):
    """
    Signal has multiple combinational drivers (this is not possible in real word)
    or signal has no driver specified and it drives something which has effect on any output of component
    (it needs to have a driver but it does not have one)

    :note: SignalDriverErr([(SignalDriverErrType, sig), ])
    """

    def __str__(self):
        try:
            ctx = self.args[0][0][1].ctx
            name = ctx.getDebugScopeName()
            scope_name = f"{name:s} of class {ctx.parent.__class__}"
        except Exception:
            scope_name = "<no parent>"

        b = [f"{self.__class__} raised in {scope_name:s}"]
        err_sigs = sorted(self.args[0], key=lambda x: (x[0].value, x[1].name))
        prev_err_t = None
        for err_t, sig in err_sigs:
            if prev_err_t is None or prev_err_t != err_t:
                b.append(SignalDriverErrType_labels[err_t])
                prev_err_t = err_t
            if err_t == SignalDriverErrType.MULTIPLE_COMB_DRIVERS or\
                    err_t == SignalDriverErrType.INPUT_WITH_DRIVER:
                b.append(f"    {sig}: {sig.drivers}")
            else:
                b.append(f"    {sig}")

        return "\n".join(b)



from typing import List

from hwt.doc_markers import internal
from hwt.hdl.statements.statement import HdlStatement
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


@internal
def HdlStatement_cut_off_drivers_of_list(sig: RtlSignalBase,
                             statements: List[HdlStatement],
                             keep_mask: List[bool],
                             new_statements: List[HdlStatement]):
    """
    Cut all logic from statements which drives signal sig.

    :param sig: signal which drivers should be removed
    :param statements: list of statements to filter
    :param keep_mask: list of flags if True statements was driver only of sig
    :param new_statements: output list of filtered statements

    :return: True if all input statements were reduced
    """
    all_cut_off = True
    for stm in statements:
        newStm = stm._cut_off_drivers_of(sig)
        keep = True
        if newStm is None:
            # statement is des not have drivers of sig
            all_cut_off = False
        elif newStm is stm:
            # statement drives only sig
            keep = False
            new_statements.append(newStm)
        else:
            # statement was splited on multiple statements
            all_cut_off = False
            new_statements.append(newStm)

        keep_mask.append(keep)

    return all_cut_off



from io import StringIO
import os
from typing import Type, Optional, Union

from hdlConvertorAst.hdlAst._bases import iHdlObj
from hdlConvertorAst.translate.common.name_scope import NameScope
from hwt.pyUtils.uniqList import UniqList
from hwt.serializer.serializer_config import DummySerializerConfig
from hwt.serializer.serializer_filter import SerializerFilter
from hwt.synthesizer.unit import HdlConstraintList
from hdlConvertorAst.hdlAst import HdlModuleDef


class StoreManager(object):
    """
    A base class for an objects which manage
    how the output of the serialization is stored by serializer_cls
    """

    def __init__(self,
                 serializer_cls: DummySerializerConfig,
                 _filter: Type["SerializerFilter"]=None,
                 name_scope: Optional[NameScope]=None):
        self.serializer_cls = serializer_cls
        self.as_hdl_ast = serializer_cls.TO_HDL_AST(name_scope=name_scope)
        self.name_scope = self.as_hdl_ast.name_scope
        if _filter is None:
            _filter = SerializerFilter()
        self.filter = _filter

    def hierarchy_push(self, obj: "Unit") -> NameScope:
        c = self.name_scope.level_push(obj.name)
        self.name_scope = c
        return c

    def hierarchy_pop(self, obj: "Unit") -> NameScope:
        p = self.name_scope.parent
        assert p is not None
        self.name_scope = p
        return p

    def write(self, obj: Union[iHdlObj, HdlConstraintList]):
        pass


class SaveToStream(StoreManager):
    """
    Store all produced code to an output stream
    """

    def __init__(self,
                 serializer_cls: DummySerializerConfig,
                 stream: StringIO,
                 _filter: "SerializerFilter"=None,
                 name_scope: Optional[NameScope]=None):
        super(SaveToStream, self).__init__(
            serializer_cls, _filter=_filter, name_scope=name_scope)
        self.stream = stream
        ser = self.ser = self.serializer_cls.TO_HDL(self.stream)
        if hasattr(ser, "stm_outputs"):
            ser.stm_outputs = self.as_hdl_ast.stm_outputs

    def write(self, obj: Union[iHdlObj, HdlConstraintList]):
        self.as_hdl_ast.name_scope = self.name_scope
        if isinstance(obj, HdlConstraintList):
            if self.serializer_cls.TO_CONSTRAINTS is not None:
                to_constr = self.serializer_cls.TO_CONSTRAINTS(self.stream)
                to_constr.visit_HdlConstraintList(obj)
        else:
            hdl = self.as_hdl_ast.as_hdl(obj)
            self.ser.visit_iHdlObj(hdl)


class SaveToFilesFlat(StoreManager):
    """
    Store all produced code to a single directory, file per component.
    """

    def __init__(self,
                 serializer_cls: DummySerializerConfig,
                 root: str,
                 _filter: "SerializerFilter"=None,
                 name_scope: Optional[NameScope]=None):
        super(SaveToFilesFlat, self).__init__(
            serializer_cls, _filter=_filter, name_scope=name_scope)
        self.root = root
        self.files = UniqList()
        self.module_path_prefix = None
        os.makedirs(root, exist_ok=True)

    def write(self, obj: Union[iHdlObj, HdlConstraintList]):
        if isinstance(obj, HdlConstraintList):
            f_name = "constraints" + self.serializer_cls.TO_CONSTRAINTS.fileExtension
        else:
            if isinstance(obj, HdlModuleDef):
                name = obj.module_name.val
            else:
                name = obj.name
            f_name = name + self.serializer_cls.fileExtension

        fp = os.path.join(self.root, f_name)
        if fp in self.files:
            m = 'a'
        else:
            m = 'w'
            self.files.append(fp)

        with open(fp, m) as f:
            s = SaveToStream(self.serializer_cls, f,
                             self.filter, self.name_scope)
            s.ser.module_path_prefix = self.module_path_prefix
            s.write(obj)



from hwt.hdl.types.hdlType import HdlType
from pyMathBitPrecise.floatt import Floatt
from hwt.doc_markers import internal


class HFloat(HdlType, Floatt):
    """
    Elemental HDL type representing IEEE 754 like float type.
    """

    def __init__(self, exponent_w, mantisa_w,
                 name=None,
                 const=False):
        """
        :param negated: if true the value is in negated form
        """
        HdlType.__init__(self, const=const)
        assert exponent_w > 0, exponent_w
        assert mantisa_w > 0, mantisa_w
        Floatt.__init__(self, exponent_w, mantisa_w, name=name)

    @internal
    @classmethod
    def getValueCls(cls):
        from hwt.hdl.types.floatVal import HFloatVal
        return HFloatVal




from collections import deque
from hwt.hdl.constants import READ, WRITE, NOP
from hwt.simulator.agentBase import SyncAgentBase
from hwtSimApi.agents.clk import ClockAgent
from hwtSimApi.hdlSimulator import HdlSimulator
from hwtSimApi.triggers import WaitCombRead, WaitWriteOnly, WaitCombStable, Timer


class BramPort_withoutClkAgent(SyncAgentBase):
    """
    A simulation agent for BramPort_withoutClk interface
    In slave mode acts as a memory, in master mode dispathes
    requests stored in "requests" dequeu

    :ivar ~.requests: list of tuples (request type, address, [write data])
        - used for driver
    :ivar ~.data: list of data in memory, used for monitor
    :ivar ~.mem: if agent is in monitor mode (= is slave) all reads and writes
        are performed on mem object
    """

    def __init__(self, sim: HdlSimulator, intf):
        super().__init__(sim, intf, allowNoReset=True)
        if intf.HAS_BE:
            raise NotImplementedError()
        self.HAS_WE = hasattr(intf, "we")
        self.requests = deque()
        self.readPending = False
        self.r_data = deque()

        self.mem = {}
        self.requireInit = True
        self.clk_ag = None

    def doReq(self, req):
        rw = req[0]
        addr = req[1]
        intf = self.intf

        if rw == READ:
            assert intf.HAS_R, intf
            rw = 0
            wdata = None
            self.readPending = True
            if self._debugOutput is not None:
                self._debugOutput.write("%s, after %r read_req: %d\n" % (
                                        self.intf._getFullName(),
                                        self.sim.now, addr))
        elif rw == WRITE:
            assert intf.HAS_W, intf
            wdata = req[2]
            rw = 1
            if self._debugOutput is not None:
                self._debugOutput.write("%s, after %r write: %d:%d\n" % (
                                        self.intf._getFullName(), self.sim.now,
                                        addr, wdata))

        else:
            raise NotImplementedError(rw)

        intf.addr.write(addr)
        if self.HAS_WE:
            intf.we.write(rw)
        if intf.HAS_W:
            intf.din.write(wdata)

    def onReadReq(self, addr):
        """
        on readReqRecieved in monitor mode
        """
        self.requests.append((READ, addr))

    def onWriteReq(self, addr, data):
        """
        on writeReqRecieved in monitor mode
        """
        self.requests.append((WRITE, addr, data))

    def monitor(self):
        """
        Handle read/write request on this interfaces

        This method is executed on clock edge.
        This means that the read data should be put on dout after clock edge.
        """
        intf = self.intf

        yield WaitCombStable()
        if self.notReset():
            en = intf.en.read()
            en = int(en)
            if en:
                if self.HAS_WE:
                    we = intf.we.read()
                    we = int(we)
                elif intf.HAS_W:
                    we = 1
                else:
                    we = 0

                addr = intf.addr.read()
                if we:
                    data = intf.din.read()
                    self.onWriteReq(addr, data)
                else:
                    self.onReadReq(addr)

        if self.requests:
            req = self.requests.popleft()
            t = req[0]
            addr = req[1]
            if t == READ:
                v = self.mem.get(addr.val, None)
                yield Timer(1)
                yield WaitWriteOnly()
                intf.dout.write(v)
            else:
                assert t == WRITE
                # yield WaitWriteOnly()
                # intf.dout.write(None)
                yield Timer(1)
                # after clock edge
                yield WaitWriteOnly()
                self.mem[addr.val] = req[2]

    def driver(self):
        intf = self.intf
        if self.requireInit:
            yield WaitWriteOnly()
            intf.en.write(0)

            if self.HAS_WE:
                intf.we.write(0)
            self.requireInit = False

        readPending = self.readPending
        yield WaitCombRead()
        if self.requests and self.notReset():
            yield WaitWriteOnly()
            req = self.requests.popleft()
            if req is NOP:
                intf.en.write(0)
                if self.HAS_WE:
                    intf.we.write(0)
                self.readPending = False
            else:
                self.doReq(req)
                intf.en.write(1)
        else:
            yield WaitWriteOnly()
            intf.en.write(0)
            if self.HAS_WE:
                intf.we.write(0)
            self.readPending = False

        if readPending:
            # in previous clock the read request was dispatched, now we are collecting the data
            yield WaitCombStable()
            # now we are after clk edge
            d = intf.dout.read()
            self.r_data.append(d)
            if self._debugOutput is not None:
                self._debugOutput.write("%s, on %r read_data: %d\n" % (
                                        self.intf._getFullName(),
                                        self.sim.now, d.val))


class BramPortAgent(BramPort_withoutClkAgent):

    def getDrivers(self):
        drivers = super(BramPortAgent, self).getDrivers()
        self.clk_ag = ClockAgent(self.sim, self.intf.clk)
        drivers.extend(self.clk_ag.getDrivers())
        return drivers



from hwt.doc_markers import internal
from hwt.hdl.operator import Operator
from hwt.hdl.operatorDefs import AllOps
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.defs import BOOL, INT
from hwt.hdl.types.slice import HSlice
from hwt.hdl.types.typeCast import toHVal
from hwt.hdl.value import HValue
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


class HStreamVal(HValue):
    """
    Class for values of HStream HDL type
    """

    @classmethod
    def from_py(cls, typeObj, val, vld_mask=None):
        """
        :param typeObj: HStream instance
        :param val: None or iterrable of values
        :param vld_mask: if is None validity is resolved from val
            if is 0 value is invalidated
            if is 1 value has to be valid
        """
        min_len = typeObj.len_min

        if vld_mask == 0:
            val = None
        element_t = typeObj.element_t
        if val is None:
            elements = [element_t.from_py(None) for _ in range(min_len)]
        else:
            elements = []
            for v in val:
                if isinstance(v, RtlSignalBase):  # is signal
                    assert v._dtype == typeObj.element_t
                    e = v
                else:
                    e = typeObj.element_t.from_py(v)
                elements.append(e)
            cur_len = len(elements)
            assert cur_len >= min_len and cur_len <= typeObj.len_max
        _mask = int(bool(val))
        if vld_mask is None:
            vld_mask = _mask
        else:
            assert (vld_mask == _mask)

        return cls(typeObj, elements, vld_mask)

    def to_py(self):
        if not self._is_full_valid():
            raise ValueError(f"Value of {self} is not fully defined")
        return [v.to_py() for v in self.val]

    @internal
    def __hash__(self):
        return hash((self._dtype, self.val, self.vld_mask))

    def _is_full_valid(self):
        return self.vld_mask == 1

    @internal
    def _getitem__val(self, key):
        """
        :atention: this will clone item from array, iterate over .val
            if you need to modify items
        """
        kv = key.val
        if not key._is_full_valid():
            raise KeyError()

        return self.val[kv].__copy__()

    def __getitem__(self, key):
        iamVal = isinstance(self, HValue)
        key = toHVal(key)
        isSLICE = isinstance(key, HSlice.getValueCls())

        if isSLICE:
            raise NotImplementedError()
        elif isinstance(key, (HValue, RtlSignalBase)):
            pass
        else:
            raise NotImplementedError(
                f"Index operation not implemented for index {key}")

        if iamVal and isinstance(key, HValue):
            return self._getitem__val(key)

        return Operator.withRes(AllOps.INDEX, [self, key], self._dtype.element_t)

    @internal
    def _setitem__val(self, index, value):
        if index._is_full_valid():
            self.val[index.val] = value.__copy__()
        else:
            self.val = {}

    def __setitem__(self, index, value):
        """
        Only syntax sugar for user, not used inside HWT

        * In HW design is not used (__getitem__ returns "reference"
            and it is used)

        * In simulator is used _setitem__val directly
        """
        if isinstance(index, int):
            index = INT.from_py(index)
        else:
            assert isinstance(self, HValue)
            assert isinstance(index._dtype, Bits), index._dtype

        if not isinstance(value, HValue):
            value = self._dtype.element_t.from_py(value)
        else:
            assert value._dtype == self._dtype.element_t, (
                value._dtype, self._dtype.element_t)

        return self._setitem__val(index, value)

    def __iter__(self):
        return iter(self.val)

    def __len__(self):
        return len(self.val)

    @internal
    def _eq__val(self, other):
        assert self._dtype.element_t == other._dtype.element_t

        eq = True
        vld = 1

        if (len(self.val) == len(other.val)):
            for a, b in zip(self.val, other.val):
                eq = eq and bool(a) == bool(b)
                if not eq:
                    break
                vld = vld & a.vld_mask & b.vld_mask
        else:
            eq = False
            vld = 0

        return BOOL.getValueCls()(BOOL, int(eq), vld)

    def _eq(self, other):
        assert isinstance(other, HStreamVal)
        return self._eq__val(other)



from hwt.interfaces.std import HandshakeSync
from hwt.interfaces.structIntf import HdlType_to_Interface
from hwt.synthesizer.param import Param


class HsStructIntf(HandshakeSync):
    """
    A handshaked interface which has a data signal of type specified in configuration of this interface
    """

    def _config(self):
        self.T = Param(None)

    def _declr(self):
        assert self.T is not None
        self.data = HdlType_to_Interface().apply(self.T)
        HandshakeSync._declr(self)



from itertools import compress
from typing import List, Set, Tuple

from hwt.doc_markers import internal
from hwt.hdl.statements.statement import HdlStatement
from hwt.hdl.statements.utils.reduction import HdlStatement_try_reduce_list
from hwt.hdl.statements.utils.signalCut import HdlStatement_cut_off_drivers_of_list
from hwt.pyUtils.uniqList import UniqList
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


class HdlStmCodeBlockContainer(HdlStatement):
    """
    Hdl block statement used also to represent a HDL process

    :ivar ~.name: name used as id in target HDL
    :ivar ~.statements: list of statements in body of process

    :note: HdlStmCodeBlockContainer do not have to be process in target HDL, for example
        simple process which contains only unconditional assignment will
        be rendered just as assignment. It depends on capabilities of the target HDL.
    """

    def __init__(self):
        super(HdlStmCodeBlockContainer, self).__init__()
        self.name = None
        self.statements = []
        self.rank = 0

    @internal
    @classmethod
    def from_known_io(cls, name: str, statements: List[HdlStatement],
                 sensitivity: Set["RtlSignal"],
                 inputs: UniqList, outputs: UniqList) -> 'HdlStmCodeBlockContainer':
        self = cls()
        self.name = name
        self.statements = statements
        self._inputs = inputs
        self._outputs = outputs
        self._sensitivity = sensitivity
        self.rank = sum(map(lambda s: s.rank, statements))
        return self

    @internal
    def _try_reduce(self) -> Tuple[List["HdlStatement"], bool]:
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._try_reduce`
        """
        new_statements, _, io_change = HdlStatement_try_reduce_list(self.statements)
        return new_statements, io_change

    @internal
    def _iter_stms(self):
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._iter_stms`
        """
        yield from self.statements


    @internal
    def _cut_off_drivers_of(self, sig: RtlSignalBase):
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._cut_off_drivers_of`
        """
        if self._sensitivity is not None or self._enclosed_for is not None:
            raise NotImplementedError(
                    "Sensitivity and enclosure has to be cleaned first")

        if len(self._outputs) == 1 and sig in self._outputs:
            # this statement has only this output, eject this statement from its parent
            self.parentStm = None  # because new parent will be asigned immediately after cutting of
            return self

        sig.drivers.discard(self)
        # try to cut off all statements which are drivers of specified signal
        # in all branches
        child_keep_mask = []

        newStatements = []
        all_cut_off = True
        all_cut_off &= HdlStatement_cut_off_drivers_of_list(
            sig, self.statements, child_keep_mask, newStatements)
        self.statements = list(compress(self.statements, child_keep_mask))

        assert not all_cut_off, "everything was cut of but this should be already known at the start"

        if newStatements:
            # parts were cut off
            # generate new statement for them
            n = self.__class__(*newStatements)

            if self.parentStm is None:
                ctx = n._get_rtl_context()
                ctx.statements.add(n)

            self._cut_off_drivers_of_regenerate_io(sig, n)

            return n

    @internal
    def _replace_child_statement(self, stm:HdlStatement,
            replacement:List[HdlStatement],
            update_io:bool) -> None:
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._replace_child_statement`
        """

        if update_io:
            raise NotImplementedError()

        i = self.statements.index(stm)

        self.rank -= stm.rank
        self.statements[i:i + 1] = replacement
        for rstm in replacement:
            rstm._set_parent_stm(self)
        # reset IO because it was shared with this statement
        stm._destroy()



"""
Sythesizer converts :class:`hwt.synthesizer.unit.Unit` instances to HDL objects.

:func:`hwt.synthesizer.utils.to_rtl` function is one of examples how to use this module.
The conversion of :class:`hwt.synthesizer.unit.Unit` instances happens mainly in
:meth:`hwt.synthesizer.unit.Unit._to_rtl` which calls other optmisation and transformations
stored in :Py:attr:`:hwt.synthesizer.unit.Unit._target_platform`.
"""



from copy import copy
from typing import Union

from hdlConvertorAst.hdlAst._defs import HdlIdDef
from hdlConvertorAst.hdlAst._expr import HdlValueId, HdlValueInt, HdlDirection
from hwt.hdl.types.array import HArray
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.defs import INT, BOOL, FLOAT64
from hwt.hdl.types.enum import HEnum
from hwt.hdl.types.float import HFloat
from hwt.hdl.types.slice import HSlice
from hwt.hdl.types.string import HString
from hwt.hdl.value import HValue
from hwt.hdl.variables import SignalItem
from hwt.serializer.exceptions import SerializerException
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


class ToHdlAst_Value():

    def is_suitable_for_const_extract(self, val: HValue):
        """
        :return: True if an value should be extracted as a constant if possible
        """
        return False

    def as_hdl_Value(self, val):
        """
        :param dst: is signal connected with value
        :param val: value object, can be instance of Signal or HValue
        """
        t = val._dtype
        if isinstance(val, RtlSignalBase):
            return self.as_hdl_SignalItem(val)

        # try to extract value as constant
        cc = self.constCache
        if cc is not None:
            if self.is_suitable_for_const_extract(val):
                c = cc.extract_const_val_as_const_var(val)
                if c is not None:
                    return self.as_hdl(c)

        if isinstance(t, HSlice):
            return self.as_hdl_HSliceVal(val)
        elif isinstance(t, HArray):
            return self.as_hdl_HArrayVal(val)
        elif isinstance(t, Bits):
            return self.as_hdl_BitsVal(val)
        elif isinstance(t, HEnum):
            return self.as_hdl_HEnumVal(val)
        elif isinstance(t, HString):
            return self.as_hdl_HStringVal(val)
        elif isinstance(t, HFloat):
            return self.as_hdl_HFloatVal(val)
        else:
            raise SerializerException(
                "can not resolve value serialization for %r"
                % (val))

    def as_hdl_int(self, val: int):
        assert isinstance(val, int), val
        return HdlValueInt(val, None, None)

    def Value_try_extract_as_const(self, val):
        return None

    def as_hdl_IntegerVal(self, val):
        return self.as_hdl_int(int(val.val))

    def as_hdl_BitsVal(self, val):
        t = val._dtype
        if t == INT:
            return self.as_hdl_IntegerVal(val)
        elif t == BOOL:
            return self.as_hdl_BoolVal(val)
        w = t.bit_length()
        return self.as_hdl_BitString(val.val, w, t.force_vector,
                                     val.vld_mask, t.signed)

    def as_hdl_HStringVal(self, val):
        return val.val

    def as_hdl_HFloatVal(self, val):
        if val._dtype != FLOAT64:
            raise NotImplementedError(val._dtype)
        return float(val)

    def as_hdl_SignalItem(self, si: Union[SignalItem, HdlIdDef],
                          declaration=False):
        if declaration:
            if isinstance(si, HdlIdDef):
                var = copy(si)
                si = si.origin
            else:
                var = HdlIdDef()
                var.name = si.name
                var.origin = si
                var.value = si._val
                var.type = si._dtype
                var.is_const = si._const
            v = var.value
            if isinstance(si, RtlSignalBase):
                if si.virtual_only:
                    var.is_latched = True
                elif si.drivers or var.direction != HdlDirection.UNKNOWN:
                    # has drivers or is port/param
                    pass
                elif si.endpoints:
                    if not v.vld_mask:
                        raise SerializerException(
                            f"Signal {si.name:s} is constant and has undefined value")
                    var.is_const = True
                else:
                    raise SerializerException(
                        f"Signal {si.name:s} should be declared but it is not used")

            if v is None:
                pass
            elif isinstance(v, RtlSignalBase):
                if v._const:
                    var.value = self.as_hdl(v)
                else:
                    # default value has to be set by reset,
                    # because it is not resolvable in compile time
                    var.value = None
                    pass
            elif isinstance(v, HValue):
                if v.vld_mask or var.is_const:
                    orig_const_cache = self.constCache
                    try:
                        self.constCache = None
                        var.value = self.as_hdl_Value(v)
                    finally:
                        self.constCache = orig_const_cache
                else:
                    # remove value if it is entirely undefined
                    var.value = None
            else:
                raise NotImplementedError(v)
            var.type = self.as_hdl_HdlType(var.type)
            return var
        else:
            if si.hidden and si.origin is not None:
                # hidden signal, render it's driver instead
                return self.as_hdl(si.origin)
            return HdlValueId(si.name, obj=si)



from typing import Union, List

from hwt.hdl.value import HValue


def isSameHVal(a: HValue, b: HValue) -> bool:
    """
    :return: True if two Value instances are same
    :note: not just equal
    """
    return a is b or (isinstance(a, HValue)
                      and isinstance(b, HValue)
                      and a.val == b.val
                      and a.vld_mask == b.vld_mask)


def areSameHVals(a: Union[None, List[HValue]],
                 b: Union[None, List[HValue]]) -> bool:
    """
    :return: True if two vectors of HValue/RtlSignal instances are same
    :note: not just equal
    """
    if a is b:
        return True
    if a is None or b is None:
        return False
    if len(a) == len(b):
        for a_, b_ in zip(a, b):
            if not isSameHVal(a_, b_):
                return False
        return True
    else:
        return False



from copy import copy
from typing import Dict, Optional, Union, List, Generator

from hdlConvertorAst.translate.common.name_scope import NameScope
from hwt.doc_markers import internal
from hwt.hdl.constants import DIRECTION, INTF_DIRECTION
from hwt.hdl.statements.assignmentContainer import HdlAssignmentContainer
from hwt.hdl.types.typeCast import toHVal
from hwt.hdl.value import HValue
from hwt.synthesizer.exceptions import IntfLvlConfErr, InterfaceStructureErr
from hwt.synthesizer.hObjList import HObjList
from hwt.synthesizer.interfaceLevel.interfaceUtils.directionFns import \
    InterfaceDirectionFns
from hwt.synthesizer.interfaceLevel.interfaceUtils.implDependent import\
    InterfaceceImplDependentFns
from hwt.synthesizer.interfaceLevel.mainBases import InterfaceBase
from hwt.synthesizer.interfaceLevel.propDeclrCollector import\
    PropDeclrCollector
from hwt.synthesizer.rtlLevel.netlist import RtlNetlist
from hwt.synthesizer.rtlLevel.rtlSignal import RtlSignal
from hwt.synthesizer.rtlLevel.utils import portItemfromSignal
from hwt.synthesizer.vectorUtils import fitTo


def _default_param_updater(self, myP, parentPval):
    myP.set_value(parentPval)


class Interface(InterfaceBase, InterfaceceImplDependentFns,
                PropDeclrCollector, InterfaceDirectionFns):
    """
    Base class for all interfaces in interface synthesizer

    :cvar _NAME_SEPARATOR: separator for nested interface names
    :ivar ~._params: [] of parameter
    :ivar ~._interfaces: [] sub interfaces
    :ivar ~._name: name assigned during synthesis
    :ivar ~._parent: parent object (Unit or Interface instance)
    :ivar ~._isExtern: If true synthesizer sets it as external port of unit
    :ivar ~._associatedClk: clock Signal (interface) associated with
        this interface if is none simulation agent try to search it on parent
    :ivar ~._associatedRst: rst(_n) Signal (interface) associated
        with this interface if is none simulation agent try to search
        it on parent

    :note: only interfaces without _interfaces have

    :ivar ~._boundedSigLvlUnit: RTL unit for which was this interface created


    Agenda of directions and HDL

    :ivar ~._masterDir: specifies which direction has this interface at master
    :ivar ~._direction: means actual direction of this interface resolved
        by its drivers
    :ivar ~._ctx: rtl netlist context of all signals and params
        on this interface after interface is registered on parent _ctx
        is merged
    :ivar ~._hdl_port: a HdlPortItem instance available once the unit is synthesized

    Agenda of simulations

    :ivar ~._ag: agent object connected to this interface
        (initialized only before simultion)
    """

    _NAME_SEPARATOR = "_"

    def __init__(self, masterDir=DIRECTION.OUT,
                 hdl_name:Optional[Union[str, Dict[str, str]]]=None,
                 loadConfig=True):
        """
        This constructor is called when constructing new interface,
        it is usually done manually while creating :class:`hwt.synthesizer.unit.Unit` or
        automatically while extracting interfaces from UnitWithSoure

        :param masterDir: direction which this interface should have for master
        :param multiplyedBy: this can be instance of integer or Param,
            this mean the interface is array of the interfaces
            where multiplyedBy is the size
        :param loadConfig: do load config in __init__
        """
        self._setAttrListener = None
        self._associatedClk = None
        self._associatedRst = None
        self._parent = None
        self._name = None

        super().__init__()
        self._masterDir = masterDir
        # Interface is instantiated inside of :class:`hwt.synthesizer.unit.Unit` first,
        # master direction actually means slave from outside view
        self._direction = INTF_DIRECTION.UNKNOWN

        self._ctx = None

        if loadConfig:
            self._loadConfig()

        # flags for better design error detection
        self._isExtern = False
        self._ag = None
        self._hdl_port = None
        self._hdl_name_override = hdl_name

    def _m(self):
        """
        Note that this interface will be master

        :return: self
        """
        assert not hasattr(self, "_interfaces") or not self._interfaces, \
            "Too late to change direction of interface"
        self._direction = DIRECTION.asIntfDirection(DIRECTION.opposite(self._masterDir))

        return self

    def __call__(self, other, exclude=None, fit=False) -> List[HdlAssignmentContainer]:
        """
        :attention: it is not call of function it is operator of assignment
        """
        assert self._direction != INTF_DIRECTION.MASTER
        try:
            return self._connectTo(other, exclude, fit)
        except Exception as e:
            # simplification of previous exception traceback
            e_simplified = copy(e)
            raise e_simplified

    @internal
    def _loadDeclarations(self):
        """
        load declaratoins from _declr method
        This function is called first for parent and then for children
        """
        if not hasattr(self, "_interfaces"):
            self._interfaces = []
        self._setAttrListener = self._declrCollector
        self._declr()
        self._setAttrListener = None

        for i in self._interfaces:
            i._loadDeclarations()
            i._setAsExtern(self._isExtern)

        if self._isExtern:
            # direction from inside of unit (reverset compared to outside direction)
            if self._direction == INTF_DIRECTION.UNKNOWN:
                self._direction = INTF_DIRECTION.MASTER
            self._setDirectionsLikeIn(self._direction)

    @internal
    def _clean(self, lockNonExternal=True):
        """
        Remove all signals from this interface (used after unit is synthesized
        and its parent is connecting its interface to this unit)
        """

        if self._interfaces:
            for i in self._interfaces:
                i._clean(lockNonExternal=lockNonExternal)

    def _connectTo(self, master, exclude=None, fit=False) -> List[HdlAssignmentContainer]:
        """
        connect to another interface interface (on rtl level)
        works like self <= master in VHDL
        """
        return list(self._connectToIter(master, exclude, fit))

    @internal
    def _connectToIter(self, master, exclude, fit) -> Generator[HdlAssignmentContainer, None, None]:
        if exclude and (self in exclude or master in exclude):
            return

        if self._interfaces:
            seen_master_intfs = []
            for ifc in self._interfaces:
                if exclude and ifc in exclude:
                    mIfc = getattr(master, ifc._name, None)
                    if mIfc is not None:
                        seen_master_intfs.append(mIfc)
                    continue

                mIfc = getattr(master, ifc._name)
                seen_master_intfs.append(mIfc)
                if exclude and mIfc in exclude:
                    continue

                if isinstance(mIfc, HValue):
                    # HStruct values
                    if (ifc._masterDir in (DIRECTION.OUT, DIRECTION.INOUT) and ifc._direction == INTF_DIRECTION.MASTER) or\
                        (ifc._masterDir == DIRECTION.IN and ifc._direction == INTF_DIRECTION.SLAVE):
                        raise IntfLvlConfErr(
                            "Invalid connection", ifc, "<=", mIfc)
                    yield from ifc._connectToIter(mIfc,
                                                  exclude,
                                                  fit)
                elif mIfc._masterDir == DIRECTION.OUT:
                    if ifc._masterDir != mIfc._masterDir:
                        raise IntfLvlConfErr(
                            "Invalid connection", ifc, "<=", mIfc)

                    yield from ifc._connectToIter(mIfc,
                                                  exclude,
                                                  fit)
                else:
                    if ifc._masterDir != mIfc._masterDir:
                        raise IntfLvlConfErr(
                            "Invalid connection", mIfc, "<=", ifc)

                    yield from mIfc._connectToIter(ifc,
                                                   exclude,
                                                   fit)
            if isinstance(master, HValue):
                master_intf_cnt = len(master._dtype.fields)
            else:
                master_intf_cnt = len(master._interfaces)

            if len(seen_master_intfs) != master_intf_cnt:
                if exclude:
                    # there is a possiblity that the master interface was excluded,
                    # but we did not see it as the interface of the same name was not present on self
                    for ifc in self._interfaces:
                        if ifc in exclude or ifc not in seen_master_intfs:
                            continue
                        else:
                            # ifc is an interface which is extra on master and is missing an equivalent on slave
                            raise InterfaceStructureErr(self, master, exclude)
                else:
                    raise InterfaceStructureErr(self, master, exclude)
        else:
            if not isinstance(master, HValue) and master._interfaces:
                raise InterfaceStructureErr(self, master, exclude)

            dstSig = toHVal(self)
            srcSig = toHVal(master)

            if fit:
                srcSig = fitTo(srcSig, dstSig)

            yield dstSig(srcSig)

    @internal
    def _signalsForInterface(self,
                             ctx: RtlNetlist,
                             res: Optional[Dict[RtlSignal, DIRECTION]],
                             name_scope: Optional[NameScope],
                             prefix='', typeTransform=None,
                             reverse_dir=False):
        """
        Generate RtlSignal _sig and HdlPortInstance _hdl_port
        for each interface which has no subinterface

        :note: if already has _sig return use it instead

        :param ctx: instance of RtlNetlist where signals should be created
        :param res: output dictionary where result should be stored
        :param prefix: name prefix for created signals
        :param name_scope: name scope used to check collisions on port names
            if this a current top (every component is checked
            when it is seen first time)
        :param typeTransform: optional function (type) returns modified type
            for signal
        """
        if self._interfaces:
            for intf in self._interfaces:
                intf._signalsForInterface(ctx, res, name_scope,
                                          prefix=prefix,
                                          typeTransform=typeTransform,
                                          reverse_dir=reverse_dir)
        else:
            assert self._sig is None, self
            t = self._dtype
            if typeTransform is not None:
                t = typeTransform(t)

            hdl_name = prefix + self._getHdlName()

            s = ctx.sig(hdl_name, t)
            s._interface = self
            self._sig = s

            if self._isExtern:
                d = INTF_DIRECTION.asDirection(self._direction)
                u = ctx.parent
                if reverse_dir:
                    d = DIRECTION.opposite(d)
                    assert self._hdl_port is None, (
                        "Now creating a hdl interface for top"
                        " it but seems that it was already created")

                if res is not None:
                    res[s] = d

                if reverse_dir:
                    pi = portItemfromSignal(s, u, d)
                    # port of current top component
                    s.name = name_scope.checked_name(s.name, s)
                    pi.connectInternSig(s)
                    ctx.ent.ports.append(pi)
                else:
                    pi = self._hdl_port
                    # port of some subcomponent which names were already checked
                    pi.connectOuterSig(s)

                self._hdl_port = pi

    def _getHdlName(self) -> str:
        """Get name in HDL """
        return HObjList._getHdlName(self)

    def _getFullName(self) -> str:
        """get all name hierarchy separated by '.' """
        return HObjList._getFullName(self)

    def _updateParamsFrom(self, otherObj, updater=_default_param_updater,
                          exclude=None, prefix=""):
        """
        :note: doc in :func:`~hwt.synthesizer.interfaceLevel.propDeclCollector._updateParamsFrom`
        """
        PropDeclrCollector._updateParamsFrom(
            self, otherObj, updater, exclude, prefix)
        return self

    def _bit_length(self) -> int:
        """Sum of all width of interfaces in this interface"""
        try:
            interfaces = self._interfaces
        except AttributeError:
            interfaces = None

        if interfaces is None:
            # not loaded interface
            _intf = self.__copy__()
            _intf._loadDeclarations()
            interfaces = _intf._interfaces

        if interfaces:
            w = 0
            for i in interfaces:
                w += i._bit_length()
            return w
        else:
            return self._dtype.bit_length()

    def __repr__(self) -> str:
        s = [self.__class__.__name__]
        s.append("name=%s" % self._getFullName())
        if hasattr(self, '_width'):
            s.append("_width=%s" % str(self._width))
        if hasattr(self, '_masterDir'):
            s.append("_masterDir=%s" % str(self._masterDir))
        return "<%s>" % (', '.join(s))



from hdlConvertorAst.hdlAst._defs import HdlIdDef
from hdlConvertorAst.to.verilog.constants import SIGNAL_TYPE
from hwt.hdl.portItem import HdlPortItem
from hwt.hdl.statements.statement import HdlStatement
from hwt.pyUtils.arrayQuery import arr_any
from hwt.synthesizer.param import Param
from ipCorePackager.constants import DIRECTION


def systemCTypeOfSig(s):
    """
    Check if is register or wire
    """
    if isinstance(s, HdlIdDef):
        s = s.origin

    if isinstance(s, HdlPortItem):
        if s.direction == DIRECTION.IN or s.direction == DIRECTION.INOUT:
            return SIGNAL_TYPE.PORT_WIRE

        t = systemCTypeOfSig(s.getInternSig())
        if t == SIGNAL_TYPE.WIRE:
            return SIGNAL_TYPE.PORT_WIRE
        elif t == SIGNAL_TYPE.REG:
            return SIGNAL_TYPE.PORT_REG
        else:
            raise ValueError(t)
    elif isinstance(s, Param):
        return SIGNAL_TYPE.PORT_REG
    elif s._const or\
        arr_any(s.drivers,
                lambda d: isinstance(d, HdlStatement)
                and d._event_dependent_from_branch is not None):

        return SIGNAL_TYPE.REG
    else:
        return SIGNAL_TYPE.WIRE



from hwt.serializer.generic.to_hdl_ast import ToHdlAst


class DummySerializerConfig():
    """
    The serializer which does not do any additional code transformations
    and does not produce any output. It is used to generate just internal representation
    of RTL code.
    """
    fileExtension = None
    TO_HDL_AST = ToHdlAst
    TO_HDL = None
    TO_CONSTRAINTS = None



from itertools import islice
from typing import List, Tuple

from hwt.doc_markers import internal
from hwt.hdl.statements.assignmentContainer import HdlAssignmentContainer
from hwt.hdl.statements.statement import HdlStatement
from hwt.pyUtils.arrayQuery import groupedby


@internal
def HdlStatement_merge_statement_lists(stmsA: List[HdlStatement], stmsB: List[HdlStatement])\
        ->List[HdlStatement]:
    """
    Merge two lists of statements into one

    :return: list of merged statements
    """
    if stmsA is None and stmsB is None:
        return None

    tmp = []

    a_it = iter(stmsA)
    b_it = iter(stmsB)

    a = None
    b = None
    a_empty = False
    b_empty = False

    while not a_empty and not b_empty:
        while not a_empty:
            a = next(a_it, None)
            if a is None:
                a_empty = True
                break
            elif a.rank == 0:
                # simple statement does not require merging
                tmp.append(a)
                a = None
            else:
                break

        while not b_empty:
            b = next(b_it, None)
            if b is None:
                b_empty = True
                break
            elif b.rank == 0:
                # simple statement does not require merging
                tmp.append(b)
                b = None
            else:
                break

        if a is not None or b is not None:
            if b is None:
                a = b
                b = None

            if a is not None and b is not None:
                a._merge_with_other_stm(b)

            tmp.append(a)
            a = None
            b = None

    return tmp


@internal
def HdlStatement_try_reduce_list(statements: List[HdlStatement]):
    """
    Simplify statements in the list
    """
    io_change = False
    new_statements = []

    for stm in statements:
        reduced, _io_change = stm._try_reduce()
        new_statements.extend(reduced)
        io_change |= _io_change

    new_statements, rank_decrease = HdlStatement_merge_statements(
        new_statements)

    new_statements, io_change, _rank_decrease = HdlStatement_reduce_overriden_assignments(new_statements)
    rank_decrease += _rank_decrease
    return new_statements, rank_decrease, io_change


@internal
def HdlStatement_reduce_overriden_assignments(statements: List[HdlStatement]):
    io_change = False
    new_statements = []
    rank_decrease = 0

    fully_driven_outputs = set()
    for stm in reversed(statements):
        if fully_driven_outputs.issuperset(stm._outputs):
            rank_decrease += stm.rank
            io_change = True
            continue

        if isinstance(stm, HdlAssignmentContainer):
            fully_driven_outputs.update(stm._outputs)

        new_statements.append(stm)

    return list(reversed(new_statements)), io_change, rank_decrease


@internal
def HdlStatement_merge_statements(statements: List[HdlStatement])\
        ->Tuple[List[HdlStatement], int]:
    """
    Merge statements in list to remove duplicated if-then-else trees

    :return: tuple (list of merged statements, rank decrease due merging)
    :note: rank decrease is sum of ranks of reduced statements
    :attention: statement list has to me mergable
    """
    order = {}
    for i, stm in enumerate(statements):
        order[stm] = i

    new_statements = []
    rank_decrease = 0

    for rank, stms in groupedby(statements, lambda s: s.rank):
        if rank == 0:
            new_statements.extend(stms)
        else:
            if len(stms) == 1:
                new_statements.extend(stms)
                continue

            # try to merge statements if they are same condition tree
            for iA, stmA in enumerate(stms):
                if stmA is None:
                    continue

                for iB, stmB in enumerate(islice(stms, iA + 1, None)):
                    if stmB is None:
                        continue

                    if stmA._is_mergable(stmB):
                        rank_decrease += stmB.rank
                        stmA._merge_with_other_stm(stmB)
                        stms[iA + 1 + iB] = None

                new_statements.append(stmA)

    new_statements.sort(key=lambda stm: order[stm])
    return new_statements, rank_decrease


@internal
def is_mergable_statement_list(stmsA, stmsB):
    """
    Walk statements and compare if they can be merged into one statement list
    """
    if stmsA is None and stmsB is None:
        return True

    elif stmsA is None or stmsB is None:
        return False

    a_it = iter(stmsA)
    b_it = iter(stmsB)

    a = _get_stm_with_branches(a_it)
    b = _get_stm_with_branches(b_it)
    while a is not None or b is not None:
        if a is None or b is None or not a._is_mergable(b):
            return False

        a = _get_stm_with_branches(a_it)
        b = _get_stm_with_branches(b_it)

    # lists are empty
    return True


@internal
def _get_stm_with_branches(stm_it):
    """
    :return: first statement with rank > 0 or None if iterator empty
    """
    last = None
    while last is None or last.rank == 0:
        try:
            last = next(stm_it)
        except StopIteration:
            last = None
            break

    return last




class NOT_SPECIFIED():
    """
    Constant which means that the thing is not specified

    Used for optional argumens as a value whchi marks that the value of this
    argument was not specified on the place where we can not just use None
    """
    def __init__(self):
        raise AssertionError("Use only a class a constant")


from hwt.simulator.agentBase import SyncAgentBase
from hwtSimApi.agents.handshaked import HandshakedAgent as pcHandshakedAgent
from hwtSimApi.hdlSimulator import HdlSimulator


class HandshakedAgent(SyncAgentBase, pcHandshakedAgent):
    """
    Simulation/verification agent for :class:`hwt.interfaces.std.Handshaked`
    interface there is onMonitorReady(simulator)
    and onDriverWriteAck(simulator) unimplemented method
    which can be used for interfaces with bi-directional data streams

    :note: 2-phase (xor) handshake

    :attention: requires clk and rst/rstn signal (
        If you do not have any create simulation wrapper with it.
        Without it you can very easily end up with a combinational loop.)
    """

    def __init__(self, sim: HdlSimulator, intf: "Handshaked", allowNoReset=False):
        rst = self._discoverReset(intf, allowNoReset)
        clk = intf._getAssociatedClk()
        pcHandshakedAgent.__init__(self, sim, intf, clk, rst)
        self._vld = self.get_valid_signal(intf)
        self._rd = self.get_ready_signal(intf)

    @classmethod
    def get_ready_signal(cls, intf):
        return intf.rd._sigInside

    def get_ready(self):
        return self._rd.read()

    def set_ready(self, val):
        self._rd.write(val)

    @classmethod
    def get_valid_signal(cls, intf):
        return intf.vld._sigInside

    def get_valid(self):
        """get "valid" signal"""
        return self._vld.read()

    def set_valid(self, val):
        return self._vld.write(val)

    def get_data(self):
        """extract data from interface"""
        return self.intf.data.read()

    def set_data(self, data):
        """write data to interface"""
        self.intf.data.write(data)


class UniversalHandshakedAgent(HandshakedAgent):
    """
    Same thing like :class:`hwt.interfaces.agents.handshaked.HandshakedAgent`
    just the get_data/set_data method is predefined to use a tuple constructed
    from signals available on this interface.

    :ivar ~._signals: tuple of data signals of this interface (excluding ready and valid signal)
    :ivar ~._sigCnt: len(_signals)
    """

    def __init__(self, sim: HdlSimulator, intf: "Handshaked", allowNoReset=False):
        HandshakedAgent.__init__(self, sim, intf, allowNoReset=allowNoReset)

        signals = []
        rd = self.get_ready_signal(intf)
        vld = self.get_valid_signal(intf)
        for i in intf._interfaces:
            if i._sigInside is not rd and i._sigInside is not vld:
                signals.append(i)
        self._signals = tuple(signals)
        self._sigCnt = len(signals)

    def get_data(self):
        if self._sigCnt == 1:
            return self._signals[0].read()
        else:
            return tuple(sig.read() for sig in self._signals)

    def set_data(self, data):
        if data is None:
            for sig in self._signals:
                sig.write(None)
        else:
            if self._sigCnt == 1:
                self._signals[0].write(data)
            else:
                assert len(data) == self._sigCnt, (
                    "invalid number of data for an interface",
                    len(data),
                    self._signals,
                    self.intf._getFullName())
                for sig, val in zip(self._signals, data):
                    try:
                        sig.write(val)
                    except:
                        raise ValueError("Error while writing ", val, "to ", sig)

class HandshakeSyncAgent(HandshakedAgent):
    """
    Simulation/verification agent for HandshakedSycn interface

    :attention: there is no data channel on this interface
        it is synchronization only and it actually does not have
        any meaningful data collected data in monitor
        mode are just values of simulation time when item was collected
    """

    def set_data(self, data):
        pass

    def get_data(self):
        return self.sim.now


class HandshakedReadListener():

    def __init__(self, hsAgent: HandshakedAgent):
        self.original_afterRead = hsAgent._afterRead
        hsAgent._afterRead = self._afterReadWrap
        self.agent = hsAgent
        self.callbacks = {}

    def _afterReadWrap(self):
        if self.original_afterRead is not None:
            self.original_afterRead()
        try:
            cb = self.callbacks.pop(len(self.agent.data))
        except KeyError:
            return
        cb()

    def register(self, transCnt, callback):
        self.callbacks[transCnt] = callback
